#!/usr/bin/env amspython
# coding: utf-8

# This example shows how to perform a geometry optimization of a water molecule and compute
# the vibrational normal modes using GFN1-xTB.
#
# If you do not have
# a DFTB license, remove the line with DFTB settings and instead set
# ``settings.input.ForceField.Type = 'UFF'``

# ## Initial imports
#
# These two lines are not needed if you run PLAMS using the ``$AMSBIN/plams`` program. They are only needed if you use ``$AMSBIN/amspython``.

from scm.plams import *

init()


# ## Initial structure

# You could also load the geometry from an xyz file:
# molecule = Molecule('path/my_molecule.xyz')
# or generate a molecule from SMILES:
# molecule = from_smiles('O')
molecule = Molecule()
molecule.add_atom(Atom(symbol="O", coords=(0, 0, 0)))
molecule.add_atom(Atom(symbol="H", coords=(1, 0, 0)))
molecule.add_atom(Atom(symbol="H", coords=(0, 1, 0)))


try:
    plot_molecule(molecule)  # plot molecule in a Jupyter Notebook in AMS2023+
except NameError:
    pass


# ## Calculation settings
#
# The calculation settings are stored in a ``Settings`` object, which is a type of nested dictionary.

settings = Settings()
settings.input.ams.Task = "GeometryOptimization"
settings.input.ams.Properties.NormalModes = "Yes"
settings.input.DFTB.Model = "GFN1-xTB"
# settings.input.ForceField.Type = 'UFF' # set this instead of DFTB if you do not have a DFTB license. You will then not be able to extract the HOMO and LUMO energies.


# ## Create an AMSJob

job = AMSJob(molecule=molecule, settings=settings, name="water_optimization")


# You can check the input to AMS by calling the ``get_input()`` method:

print("-- input to the job --")
print(job.get_input())
print("-- end of input --")


# ## Run the job

job.run()


# ## Main results files: ams.rkf and dftb.rkf
#
# The paths to the main binary results files ``ams.rkf`` and ``dftb.rkf`` can be retrieved as follows:

print(job.results.rkfpath(file="ams"))
print(job.results.rkfpath(file="engine"))


# ## Optimized coordinates

optimized_molecule = job.results.get_main_molecule()

print("Optimized coordinates")
print("---------------------")
print(optimized_molecule)
print("---------------------")


try:
    plot_molecule(optimized_molecule)  # plot molecule in a Jupyter Notebook in AMS2023+
except NameError:
    pass


# ## Optimized bond lengths and angle

# Unlike python lists, where the index of the first element is 0,
# the index of the first atom in the molecule object is 1.

bond_length = optimized_molecule[1].distance_to(optimized_molecule[2])
print("O-H bond length: {:.3f} angstrom".format(bond_length))


bond_angle = optimized_molecule[1].angle(optimized_molecule[2], optimized_molecule[3])
print("Bond angle  : {:.1f} degrees".format(Units.convert(bond_angle, "rad", "degree")))


# ## Calculation timing

timings = job.results.get_timings()

print("Timings")
print("-------")
for key, value in timings.items():
    print(f"{key:<20s}: {value:.3f} seconds")
print("-------")


# ## Energy

energy = job.results.get_energy(unit="kcal/mol")

print("Energy      : {:.3f} kcal/mol".format(energy))


# ## Vibrational frequencies

frequencies = job.results.get_frequencies(unit="cm^-1")

print("Frequencies")
print("-----------")
for freq in frequencies:
    print(f"{freq:.3f} cm^-1")
print("-----------")


# ##  Dipole moment

import numpy as np

try:
    dipole_moment = np.linalg.norm(np.array(job.results.get_dipolemoment()))
    dipole_moment *= Units.convert(1.0, "au", "debye")
    print("Dipole moment: {:.3f} debye".format(dipole_moment))
except KeyError:
    print("Couldn't extract the dipole moment")


# ## HOMO, LUMO, and HOMO-LUMO gap
#
# Note: The methods for extracting HOMO, LUMO, and HOMO-LUMO gap only exist in AMS2023 and later.

try:
    homo = job.results.get_homo_energies(unit="eV")[0]
    lumo = job.results.get_lumo_energies(unit="eV")[0]
    homo_lumo_gap = job.results.get_smallest_homo_lumo_gap(unit="eV")

    print("HOMO        : {:.3f} eV".format(homo))
    print("LUMO        : {:.3f} eV".format(lumo))
    print("HOMO-LUMO gap : {:.3f} eV".format(homo_lumo_gap))
except AttributeError:
    print("Methods to extract HOMO and LUMO require AMS2023 or later")
except KeyError:
    print("Couldn't extract the HOMO and LUMO.")


# ## Read results directly from binary .rkf files
#
# You can also read results directly from the binary .rkf files. Use the "expert mode" of the KFbrowser program that comes with AMS to find out which section and variable to read.
#
# Below, we show how to extract the ``AMSResults%Energy`` variable from the dftb.rkf file. This is the same number that was extracted previously using the ``job.results.get_energy()`` method.

energy = job.results.readrkf("AMSResults", "Energy", file="engine")
print(f"Energy from the engine .rkf file (in hartree): {energy}")


# ## Finish PLAMS
#
# The ``finish()`` method is called automatically if you run the script with ``$AMSBIN/plams``. You should only call it if you use ``$AMSBIN/amspython`` to run the script.

finish()



#!/usr/bin/env amspython
# coding: utf-8

# ## Initial imports

from scm.plams import *

try:
    from ase import Atoms
    from pymatgen.core.structure import Structure
    from pymatgen.analysis.diffraction.xrd import XRDCalculator
except ImportError as e:
    print(
        "You need ASE and pymatgen installed in the AMS python environment to run this example. Install the package for m3gnet to do this."
    )
    print(e)
    exit(1)


# ## Create ASE atoms object for BaTiO3

at = Atoms(
    symbols=[
        "Ba",
        "Ti",
        "O",
        "O",
        "O",
    ],
    scaled_positions=[
        [
            0.0,
            0.0,
            0.0,
        ],
        [0.5, 0.5, 0.5],
        [0.0, 0.0, 0.5],
        [0.0, 0.5, 0.0],
        [0.5, 0.0, 0.0],
    ],
    cell=[4.01, 4.01, 4.01],
    pbc=(True, True, True),
)
plot_molecule(at, rotation="-5x,5y,0z")  # show in Jupyter notebook


# ## Save ASE Atoms to .cif format

fname = "batio3.cif"
at.write(fname)


# ## Load .cif in pymatgen and calculate XRD

# Available radiation sources:

print(f"Available radiation sources: {XRDCalculator.AVAILABLE_RADIATION}")


# Let's choose Cu K-alpha (default):

structure = Structure.from_file(fname)
xrd_calc = XRDCalculator(wavelength="CuKa")
xrd_calc.show_plot(structure)


pattern = xrd_calc.get_pattern(structure)
print("2*Theta Intensity hkl d_hkl(angstrom)")
for two_theta, intensity, hkls, d_hkl in zip(pattern.x, pattern.y, pattern.hkls, pattern.d_hkls):
    hkl_tuples = [hkl["hkl"] for hkl in hkls]
    for hkl in hkl_tuples:
        label = ", ".join(map(str, hkl))
        print(f"{two_theta:.2f} {intensity:.2f} {hkl} {d_hkl:.3f}")



from collections import OrderedDict
from itertools import product

from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results
import numpy as np
from scm.plams.tools.units import Units

__all__ = ["NumHessJob", "NumHessResults"]  # names exported to the main namespace


class NumHessResults(Results):

    def get_hessian(self, mass_weighted=False):
        j = self.job
        n = len(j.molecule)
        hessian = np.empty((3 * n, 3 * n))

        for atom, axis in product(range(1, n + 1), range(3)):
            v1 = np.array(j.gradient(j.children[(atom, axis, -1)].results))
            v2 = np.array(j.gradient(j.children[(atom, axis, 1)].results))
            hessian[3 * (atom - 1) + axis] = (v2 - v1) / (2 * Units.convert(j.step, "angstrom", "bohr"))

        if mass_weighted:
            masses = [at.mass for at in j.molecule]
            weights = np.empty((n, n))
            for i, j in product(range(n), repeat=2):
                weights[i, j] = masses[i] * masses[j]
            hessian /= np.kron(weights, np.ones((3, 3)))
        return (hessian + hessian.T) / 2


class NumHessJob(MultiJob):
    """A class for calculating numerical hessian.

    The length and unit of the geometry displacement step can be adjusted.
    *gradient* should be a function that takes results of *jobtype* and
    returns energy gradient in hartree/bohr.
    """

    _result_type = NumHessResults

    def __init__(self, molecule, step=0.01, unit="angstrom", jobtype=None, gradient=None, **kwargs):
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)
        self.molecule = molecule
        self.step = Units.convert(step, unit, "angstrom")
        self.jobtype = jobtype  # who is going to calculate single points
        self.gradient = gradient  # function extracting gradients from children

    def prerun(self):  # noqa F811
        for atom, axis, step in product(range(1, 1 + len(self.molecule)), range(3), [-1, 1]):
            vec = [0, 0, 0]
            vec[axis] = self.step * step
            newmol = self.molecule.copy()
            newmol[atom].translate(vec)
            newname = "{}_{}_{}".format(atom, axis, step)
            self.children[(atom, axis, step)] = self.jobtype(name=newname, molecule=newmol, settings=self.settings)



import os
import json
import threading

from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.amsworker import AMSWorker, AMSWorkerError

__all__ = []

# !!!!!!!  DEPRECATED  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# The class in this module has been deprecated. The equivalent in scm.libbase
# should be used instead where possible. The scm.libbase version does the input
# parsing via direct calls into libscm_base, instead of spawning an AMSWorker
# and then pushing the input through the pipe. This implementation exists here
# only to remove the dependency on scm.libbase when running in python
# environments without access to the base library.
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


class InputParser:
    """
    A utility class for converting text input into JSON dictionaries and plams.Settings.

    This is a legacy implementation for environments without access to scm.libbase.
    """

    def __init__(self):
        sett = Settings()
        sett.input.LennardJones = Settings()
        sett.runscript.nproc = 1
        sett.runscript.preamble_lines = [f'export AMS_INPUTREADER_ROOTPATH="{os.getcwd()}"']
        self.worker = AMSWorker(sett)

    def __enter__(self):
        return self

    def stop(self, keep_workerdir=False):
        if self.worker is not None:
            self.worker.stop(keep_workerdir)

    def __exit__(self, *args):
        self.stop()

    def _run(self, program, text_input, string_leafs=True):
        """Run a string of text input through the input parser and produce a Python dictionary representing the JSONified input."""
        try:
            json_input = self.worker.ParseInput(program, text_input, string_leafs)
        except AMSWorkerError as exc:
            raise ValueError(f"Input parsing failed. {exc.get_errormsg()}") from exc
        return json.loads(json_input)

    def to_settings(self, program, text_input):
        """Transform a string with text input into a PLAMS Settings object."""

        if program in ["ams", "acerxn"]:
            input_settings = Settings()
            lines = text_input.splitlines()

            # Find the lines corresponding to the engine block.
            while 1:
                lines, engine_lines = self.separate_engine_lines(lines)
                if engine_lines is None:
                    break
                # We have found a separate engine block.
                engine_name = engine_lines[0].split()[1].lower()
                if len(engine_lines) == 2:
                    # If it's empty we already know the result of parsing it.
                    input_settings[engine_name] = Settings()
                else:
                    input_settings[engine_name] = Settings(self._run(engine_name, "\n".join(engine_lines[1:-1])))

            input_settings["ams"] = Settings(self._run(program, "\n".join(lines)))

        else:
            input_settings = Settings(self._run(program, text_input))

        return input_settings

    @staticmethod
    def separate_engine_lines(lines):
        """
        Separate the engine lines from other lines of AMS input text
        """
        # Find the lines corresponding to the engine block.
        engine_lines = None
        engine_first, engine_last = None, None
        inner_engines = 0
        ends = 0
        for i, l in enumerate(lines):
            if l == "" or l.isspace():
                continue
            if l.split(maxsplit=1)[0].lower() == "engine":
                if engine_first is None:
                    engine_first = i
                    continue
                else:
                    inner_engines += 1
            elif l.split(maxsplit=1)[0].lower() == "endengine":
                ends += 1
            if ends > inner_engines:
                engine_last = i
                break
        if engine_first is not None and engine_last is not None:
            # We have found a separate engine block.
            engine_lines = lines[engine_first : engine_last + 1]
            # We have dealt with the engine lines, let's remove them from the input.
            del lines[engine_first : engine_last + 1]
        return lines, engine_lines


class InputParserFacade:
    """
    A utility class for converting text input into JSON dictionaries and plams.Settings.

    Uses the scm.libbase implementation of InputParser if available, and otherwise the legacy implementation
    of the parser which spawns an AMSWorker instance.
    """

    try:
        from scm.libbase import InputParser as InputParserScmLibbase

        # Cache a single instance of the parser to avoid having to repeatedly reload input file definition JSON
        # But for this need to make access to the parser thread-safe
        input_parser_scm_libbase = InputParserScmLibbase()
        input_parser_lock = threading.Lock()
        _has_scm_libbase = True
    except ImportError:
        _has_scm_libbase = False

    @property
    def parser(self):
        """
        Get instance of a parser used to convert text input.
        """
        if self._has_scm_libbase:
            return self.input_parser_scm_libbase
        else:
            return InputParser()

    def to_settings(self, program: str, text_input: str):
        """Transform a string with text input into a PLAMS Settings object."""

        if self._has_scm_libbase:
            with self.input_parser_lock:
                return self.parser.to_settings(program, text_input)
        else:
            with self.parser as parser:
                return parser.to_settings(program, text_input)

    def to_dict(self, program: str, text_input: str, string_leafs: bool = True):
        """
        Run a string of text input through the input parser and produce
        a Python dictionary representing the JSONified input.
        """
        if self._has_scm_libbase:
            with self.input_parser_lock:
                return self.parser.to_dict(program, text_input, string_leafs)
        else:
            with self.parser as parser:
                return parser._run(program, text_input, string_leafs)



"""Implementation of the AMSPipeCalculator class.

"""

from copy import deepcopy

import numpy as np

from scm.plams.core.functions import config
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.amsworker import AMSWorker
from scm.plams.interfaces.molecule.ase import fromASE, toASE

__all__ = ["AMSCalculator", "BasePropertyExtractor"]

try:
    from ase.calculators.calculator import Calculator, all_changes
    from ase.units import Bohr, Hartree
except ImportError:
    # empty interface if ase does not exist:
    __all__ = []

    class Calculator:
        def __init__(self, *args, **kwargs):
            raise NotImplementedError("AMSCalculator can not be used without ASE")

    all_changes = []
    Hartree = Bohr = 0


class BasePropertyExtractor:
    def __call__(self, ams_results, atoms):
        return self.extract(ams_results, atoms)

    def extract(self, ams_result, atoms):
        raise NotImplementedError

    def set_settings(self, settings):
        pass

    def check_settings(self, settings):
        return True


class EnergyExtractor(BasePropertyExtractor):
    name = "energy"

    def extract(self, ams_results, atoms):
        return ams_results.get_energy() * Hartree


def canonicalize_string(possible_string):
    try:
        return possible_string.lower().strip()
    except (AttributeError, TypeError):
        return possible_string


def is_ams_true(value):
    return canonicalize_string(value) in [True, "true", "yes"]


def is_ams_false(value):
    return not is_ams_true(value)


class ForceExtractor(BasePropertyExtractor):
    name = "forces"

    def extract(self, ams_results, atoms):
        return -ams_results.get_gradients() * Hartree / Bohr

    def set_settings(self, settings):
        settings.input.ams.Properties.Gradients = "Yes"
        return settings

    def check_settings(self, settings):
        return is_ams_true(settings.copy().input.ams.Properties.Gradients)


class StressExtractor(BasePropertyExtractor):
    name = "stress"

    def extract(self, ams_results, atoms):
        D = sum(atoms.get_pbc())
        if isinstance(atoms.get_pbc(), bool):
            D *= 3
        st = ams_results.get_stresstensor() * Hartree / Bohr**D
        xx, yy, zz, yz, xz, xy = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
        if D >= 1:
            xx = st[0][0]
            if D >= 2:
                yy = st[1][1]
                xy = st[0][1]
                if D >= 3:
                    zz = st[2][2]
                    yz = st[1][2]
                    xz = st[0][2]
        return np.array([xx, yy, zz, yz, xz, xy])

    def set_settings(self, settings):
        settings.input.ams.Properties.StressTensor = "Yes"
        return settings

    def check_settings(self, settings):
        return is_ams_true(settings.copy().input.ams.Properties.StressTensor)


class AMSCalculator(Calculator):
    """
    ASE Calculator which can run any AMS engine (ADF, BAND, DFTB, ReaxFF, MLPotential, ForceField, ...).

    The settings are specified with a PLAMS ``Settings`` object in the same way as when running AMS through PLAMS.

    .. important::

        Before initializing the AMSCalculator you need to call ``plams.init()``:

        .. code-block:: python

            from scm.plams import *
            init()


    Parameters:

    settings  : Settings
                A Settings object representing the input for an AMSJob or AMSWorker.
                This also determines which `implemented_properties` are available:
                `settings.input.ams.properties.gradients`: `force`
                `settings.input.ams.properties.stresstensor`: `stress`
    name      : str, optional
                Name of the rundir of calculations done by this calculator. A counter
                is appended to the name for every calculation.
    amsworker : bool , optional
                If True, use the AMSWorker to set up an interactive session.
                The AMSWorker will spawn a seperate
                process (an amsdriver). In order to make sure this process is closed,
                either use AMSCalculator as a context manager or ensure that
                AMSCalculator.stop_worker() is called before python is finished:

                .. code-block:: python

                    with AMSCalculator(settings=settings, amsworker=True) as calc:
                        atoms.set_calculator(calc)
                        atoms.get_potential_energy()

                If False, use AMSJob to set up an io session (a normal AMS calculation storing all output on disk).
    restart   : bool , optional
                Allow the engine to restart based on previous calculations.
    molecule  : Molecule , optional
                A Molecule object for which the calculation has to be performed. If
                settings.input.ams.system is defined it overrides the molecule argument.
                If AMSCalculator.calculate(atoms = atoms) is called with an atoms argument
                it overrides any earlier definition of the system and remembers it.
    extractors: List[BasePropertyExtractor] , optional
                Define extractors for additional properties.


    Examples:
    """

    # counters are a dict as a class variable. This is to support deepcopying/multiple instances with the same name
    _counter = {}

    def __new__(cls, settings=None, name="", amsworker=False, restart=True, molecule=None, extractors=[]):
        """Dispatch object creation to AMSPipeCalculator or AMSJobCalculator depending on |amsworker|"""
        if cls == AMSCalculator:
            if amsworker:
                obj = object.__new__(AMSPipeCalculator)
            else:
                obj = object.__new__(AMSJobCalculator)
        else:
            obj = object.__new__(cls)
        return obj

    def __init__(self, settings=None, name="", amsworker=False, restart=True, molecule=None, extractors=[]):

        if not isinstance(settings, Settings):
            settings = Settings.from_dict(settings)
        else:
            settings = settings.copy()

        self.settings = settings.copy()
        self.amsworker = amsworker
        self._name = name
        self.restart = restart
        self.molecule = molecule
        extractors = settings.pop("Extractors", [])
        self.extractors = [EnergyExtractor(), ForceExtractor(), StressExtractor()]
        self.extractors += [e for e in extractors if not e in self.extractors]

        if "system" in self.settings.input.ams:
            mol_dict = AMSJob.settings_to_mol(settings)
            atoms = toASE(mol_dict[""]) if mol_dict else None
        elif molecule:
            atoms = toASE(molecule)
        else:
            atoms = None

        super().__init__()
        self.atoms = atoms

        self.prev_ams_results = None
        self.results = dict()
        self.properties_updated = False

    @property
    def counter(self):
        # this is needed for deepcopy/pickling etc
        if self.name not in self._counter:
            self.set_counter()
        self._counter[self.name] += 1
        return self._counter[self.name]

    def set_counter(self, value=0):
        self._counter[self.name] = value

    @property
    def implemented_properties(self):
        """Returns the list of properties that this calculator has implemented"""
        return [extractor.name for extractor in self.extractors if extractor.check_settings(self.settings)]

    @property
    def name(self):
        return self._name

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """Calculate the requested properties. If atoms is not set, it will reuse the last known Atoms object."""
        if atoms is not None:
            # no need to redo the calculation, we already have everything.
            if (
                self.atoms == atoms
                and system_changes == []
                and all([p in self.results for p in properties])
                and not self.properties_updated
            ):
                return
            self.atoms = atoms.copy()
        self.properties_updated = False
        if self.atoms is None:
            raise ValueError("No atoms object was set.")

        if not config.init:
            raise RuntimeError("Before AMSCalculator can calculate results you need to call plams.init()")

        molecule = fromASE(self.atoms, set_charge=True)
        ams_results = self._get_ams_results(molecule, properties)
        if not ams_results.ok():
            self.results = dict()
            return

        self.results_from_ams_results(ams_results, self._get_job_settings(properties))
        self.prev_ams_results = ams_results

    def ensure_property(self, properties):
        """A list of ASE properties that the calculator will ensure are available from AMS or it gives an error."""
        if isinstance(properties, str):
            properties = [properties]
        for prop in properties:
            property_found = False
            for extractor in self.extractors:
                if prop == extractor.name:
                    self.settings = extractor.set_settings(self.settings.copy())
                    self.properties_updated = True
                property_found = True
            if not property_found:
                raise NotImplementedError(f"No extractor known for property {prop}")

    def results_from_ams_results(self, ams_results, job_settings):
        """Populates the self.results dictionary by having extractors act on an AMSResults object."""
        for extractor in self.extractors:
            if extractor.check_settings(job_settings):
                self.results[extractor.name] = extractor.extract(ams_results, self.atoms)

    def _get_ams_results(self, molecule, properties):
        raise NotImplementedError("Subclasses of AMSCalculator should implement this")

    def _get_job_settings(self, properties):
        """Returns a Settings object which ensures that an AMS calculation is run from which all requested
        properties can be extracted"""
        return self.settings.copy()

    def stop_worker(self):
        """Stops the amsworker if it exists"""
        if hasattr(self, "worker") and self.worker:
            stop = self.worker.stop()
            self.worker = None
            return stop
        else:
            # this is what AMSWorker.stop() would return if it was already stopped previously
            self.worker = None
            return (None, None)

    def clean_exit(self):
        """Function called by ASEPipeWorker to tell the Calculator to stop and clean up"""
        self.stop_worker()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.stop_worker()

    @property
    def amsresults(self):
        if hasattr(self, "prev_ams_results"):
            return self.prev_ams_results


class AMSPipeCalculator(AMSCalculator):
    """This class should be instantiated through AMSCalculator with settings.Calculator.Pipe defined"""

    def __init__(self, settings=None, name="", amsworker=True, restart=True, molecule=None, extractors=[]):
        super().__init__(settings, name, amsworker, restart, molecule, extractors)

        self.worker_settings = self.settings.copy()
        if "Task" in self.worker_settings.input.ams:
            del self.worker_settings.input.ams.Task
        if "Properties" in self.worker_settings.input.ams:
            del self.worker_settings.input.ams.Properties
        self.worker = None

    def _get_ams_results(self, molecule, properties):
        job_settings = self._get_job_settings(properties)
        if self.worker is None:
            self.worker = AMSWorker(self.worker_settings, use_restart_cache=self.restart)
        # AMSWorker expects no engine definition at this point.
        s = Settings()
        s.input.ams = job_settings.input.ams
        if "amsworker" in job_settings:
            s.amsworker = job_settings.amsworker
        s.amsworker.prev_results = self.prev_ams_results
        job_settings = s
        # run the worker from _solve_from_settings
        return self.worker._solve_from_settings(
            name=self.name + str(self.counter), molecule=molecule, settings=job_settings
        )

    def __deepcopy__(self, memo):
        """The AMSWorker instance is not copied, but instead, all the copies use the same worker"""
        memo[id(self.worker)] = self.worker
        try:
            this_method = self.__deepcopy__
            self.__deepcopy__ = None
            copy = deepcopy(self, memo)
            self.__deepcopy__ = this_method
            return copy
        except Exception as e:
            self.__deepcopy__ = this_method
            raise e


class AMSJobCalculator(AMSCalculator):
    """This class should be instantiated through AMSCalculator with settings.Calculator.AMSJob defined"""

    def _get_ams_results(self, molecule, properties):
        job_settings = self._get_job_settings(properties)
        if self.restart and self.prev_ams_results:
            job_settings.input.ams.EngineRestart = self.prev_ams_results.rkfpath(file="engine")

        return AMSJob(name=self.name + str(self.counter), molecule=molecule, settings=job_settings).run()



#!/usr/bin/env amspython
import os

import matplotlib.pyplot as plt
from scm.plams import Settings, Units, from_smiles, CRSJob, init, finish
from scm.plams.recipes.adfcosmorscompound import ADFCOSMORSCompoundJob


def solubility():
    # database can also be replaced with the output of "$AMSBIN/amspackages loc adfcrs" /ADFCRS-2018
    database = CRSJob.database()
    if not os.path.exists(database):
        raise OSError(f"The provided path does not exist. Exiting.")

    solute_smiles = "c1ccccc1"
    solute_coskf = generate_coskf(solute_smiles, "adf_benzene")  # generate files with ADF
    # solute_coskf = os.path.abspath('plams_workdir/adf_benzene/adf_benzene.coskf') # to not rerun the ADF calculation
    # solute_coskf = os.path.join(database, 'Benzene.coskf') # to load from database

    # You can also estimate the solute properties with the Property Prediction tool. See the Property Prediction example
    solute_properties = {"meltingpoint": 278.7, "hfusion": 9.91}  # experimental values for benzene, hfusion in kJ/mol

    solvent_coskf = os.path.join(database, "Water.coskf")
    solvent_density = 1.0

    s = Settings()
    s.input.property._h = "solubility"
    s.input.property.DensitySolvent = solvent_density
    s.input.temperature = "273.15 283.15 10"
    s.input.pressure = "1.01325 1.01325 10"

    s.input.compound = [Settings(), Settings()]

    s.input.compound[0]._h = solvent_coskf
    s.input.compound[0].frac1 = 1.0

    s.input.compound[1]._h = solute_coskf
    s.input.compound[1].nring = 6  # number of ring atoms benzene
    s.input.compound[1].meltingpoint = solute_properties["meltingpoint"]
    s.input.compound[1].hfusion = solute_properties["hfusion"] * Units.convert(
        1.0, "kJ/mol", "kcal/mol"
    )  # convert from kJ/mol to kcal/mol

    job = CRSJob(name="benzene_in_water", settings=s)
    job.run()

    plot_results(job)


def generate_coskf(smiles, jobname=None):
    molecule = from_smiles(smiles, nconfs=100, forcefield="uff")[0]
    job = ADFCOSMORSCompoundJob(name=jobname, molecule=molecule)
    job.run()
    plot_sigma_profile(job.results)
    return job.results.coskfpath()


def plot_results(job):
    res = job.results.get_results("SOLUBILITY")
    solubility_g_L = res["solubility g_per_L_solvent"][1]
    temperatures = res["temperature"]
    for temperature, sol_g_l in zip(temperatures, solubility_g_L):
        print(f"{temperature:.2f} {sol_g_l:.4f}")

    plt.plot(temperatures, solubility_g_L)
    plt.xlabel("Temperature (K)")
    plt.ylabel("Solubility (g/L solvent)")
    plt.show()


def get_sigma_profile(coskf_file):
    s = Settings()
    s.input.property._h = "PURESIGMAPROFILE"
    s.input.compound._h = coskf_file
    job = CRSJob(name="sigma_profile", settings=s)
    res = job.run()
    return res.get_sigma_profile()


def plot_sigma_profile(results):
    coskf_path = results.coskfpath()
    sigma = get_sigma_profile(coskf_path)
    xlabel = "σ (e/A**2)"
    for profile in sigma:
        if profile == xlabel:
            continue
        plt.plot(sigma[xlabel], sigma[profile], label=profile.split(".")[0])

    plt.xlabel("σ (e/Å**2)")
    plt.ylabel("p(σ)")
    plt.legend()
    plt.show()


def main():
    solubility()


if __name__ == "__main__":
    init()
    main()
    finish()



import numpy as np

from scm.plams.core.functions import config, delete_job, finish, init
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob
from scm.plams.interfaces.adfsuite.amsworker import AMSWorker
from scm.plams.mol.molecule import Molecule

__all__ = ["preoptimize", "refine_density", "refine_lattice"]


def preoptimize(
    molecule: Molecule, model: str = "UFF", settings: Settings = None, nproc: int = 1, maxiterations: int = 100
):
    """
    Returns an optimized Molecule (or list of optimized molecules)

    molecule: Molecule or list of Molecules
        Molecule to optimize
    model: str
        Shorthand for some model, e.g. 'UFF'
    settings: Settings
        Custom engine settings (overrides ``model``)
    nproc: int
        Number of processes
    maxiterations: int
        Maximum number of iterations for the geometry optimization.
    """
    my_settings = _get_quick_settings(model, settings, nproc)
    single_molecule = isinstance(molecule, Molecule)

    input_molecules = [molecule] if single_molecule else molecule
    output_molecules = []
    with AMSWorker(my_settings) as worker:
        for i, mol in enumerate(input_molecules):
            results = worker.GeometryOptimization(
                name=f"preoptimization_{i}", molecule=mol, maxiterations=maxiterations, pretendconverged=True
            )
            output_molecules.append(results.get_main_molecule())

    if single_molecule:
        return output_molecules[0]
    else:
        return output_molecules


def refine_density(
    molecule: Molecule,
    density: float,
    step_size=50,
    model: str = "UFF",
    settings: Settings = None,
    nproc: int = 1,
    maxiterations: int = 100,
):
    """

    Performs a series of geometry optimizations with densities approaching
    ``density``. This can be useful if you want to compress a system to a
    given density, but cannot just use apply_strain() (because
    apply_strain() also scales the bond lengths).

    This function can be useful if for example packmol does not succeed to
    pack molecules with the desired density. Packmol can then generate a
    structure with a lower density, and this function can be used to
    increase the density to the desired value.

    Returns: a Molecule with the requested density.

    molecule: Molecule
        The molecule must have a 3D lattice
    density: float
        Target density in kg/m^3   (1000x the density in g/cm^3)
    step_size: float
        Step size for the density (in kg/m^3). Set step_size to a large number to only use 1 step.
    model: str
        e.g. 'UFF'
    settings: Settings
        Engine settings (overrides ``model``)
    maxiterations: int
        maximum number of iterations for the geometry optimization.


    """

    assert len(molecule.lattice) == 3
    tolerance = 1e-3
    my_settings = _get_quick_settings(model, settings, nproc)
    current_density = molecule.get_density()  # kg/m^3
    output_molecule = molecule.copy()
    counter = 0
    with AMSWorker(my_settings) as worker:
        while current_density < density - tolerance or current_density > density + tolerance:
            counter += 1
            if current_density < density - step_size:
                new_density = current_density + step_size
            elif current_density > density + step_size:
                new_density = current_density - step_size
            else:
                new_density = density

            output_molecule.set_density(new_density)

            results = worker.GeometryOptimization(
                name=f"preoptimization_{counter}",
                molecule=output_molecule,
                maxiterations=maxiterations,
                pretendconverged=True,
            )

            output_molecule = results.get_main_molecule()
            current_density = output_molecule.get_density()  # kg/m^3

    return output_molecule


def refine_lattice(
    molecule: Molecule,
    lattice,
    n_points=None,
    max_strain=0.15,
    model: str = "UFF",
    settings: Settings = None,
    nproc: int = 1,
    maxiterations: int = 10,
):
    """

    Returns a ``Molecule`` for which the lattice of the ``molecule`` is
    transformed to ``lattice``, by performing short geometry optimizations
    (each for at most ``maxiterations``) on gradually distorted lattices
    (linearly interpolating from the original lattice to the new lattice
    using ``n_points`` points).

    This can be useful for transforming an orthorhombic box of a liquid
    into a non-orthorhombic box of a liquid, where the gradual transformation
    of the lattice ensures that the molecules do not become too distorted.

    If init() has been called before calling this function, the job will be run in
    the current PLAMS working directory and will be deleted when the job finishes.

    Returns: a Molecule with the requested lattice. If the refinement
    fails, ``None`` is returned.

    molecule: Molecule
        The initial molecule
    lattice: list of list of float
        List with 1, 2, or 3 elements. Each element is a list of float with 3 elements each. For example, ``lattice=[[10, 0, 0],[-5, 5, 0],[0, 0, 12]]``.
    n_points: None or int >=2
        Number of points used for the linear interpolation. If None, n_points will be chosen such that the maximum strain for any step is at most ``max_strain`` compared to the original lattice vector lengths.
    max_strain: float
        Only if ``n_points=None``, use this value to determine the maximum allowed strain from one step to the next (as a fraction of the length of the original lattice vectors).
    model: str
        e.g. 'UFF'
    settings: Settings
        Engine settings (overrides ``model``)
    nproc: int
        Number of processes used by the job
    maxiterations: int
        maximum number of iterations for the geometry optimizations.

    """
    assert len(lattice) == len(
        molecule.lattice
    ), f"Different number of lattice vectors: len(molecule.lattice) = {len(molecule.lattice)}, len(lattice) = {len(lattice)}"
    assert all(len(x) == 3 for x in lattice), f"Lattice vectors must have three components. Lattice: {lattice}"
    assert all(len(x) == 3 for x in molecule.lattice), f"Lattice vectors must have three components. Lattice: {lattice}"
    assert (
        len(lattice) >= 1 and len(lattice) <= 3
    ), f"{len(lattice)} lattice vectors given but must be between 1 and 3. Lattice: {lattice}"

    def lattice2str(latt):
        return "\n".join(" ".join(str(j) for j in i) for i in latt)

    if n_points is None:
        lat1 = np.array(molecule.lattice)
        lat2 = np.array(lattice)
        strain = np.linalg.norm(lat2, axis=1) / np.linalg.norm(lat1, axis=1)
        n_points = np.ceil(np.max(np.abs(strain - 1) / max_strain))
        n_points = int(max(n_points, 2))

    assert n_points >= 2, f"Expected n_points >=2, but received {n_points}"

    called_plams_init = _ensure_init()

    s = _get_quick_settings(model, settings, nproc)
    s.input.ams.task = "PESScan"
    s.input.ams.PESScan.ScanCoordinate.FromLattice._1 = lattice2str(molecule.lattice)
    s.input.ams.PESScan.ScanCoordinate.ToLattice._1 = lattice2str(lattice)
    s.input.ams.PESScan.ScanCoordinate.NPoints = n_points
    s.input.ams.GeometryOptimization.MaxIterations = maxiterations
    s.input.ams.GeometryOptimization.PretendConverged = "Yes"

    job = AMSJob(settings=s, molecule=molecule, name="refine_density")
    job.run()

    final_molecule = None
    if job.ok():
        results = job.results.get_pesscan_results(molecules=True)
        final_molecule = results["Molecules"][-1]

    delete_job(job)

    if called_plams_init:
        finish()

    return final_molecule


def _ensure_init():
    if config.init:
        called_plams_init = False
    else:
        s = Settings()
        s.erase_workdir = True
        s.log.stdout = 0
        init(config_settings=s)
        called_plams_init = True

    return called_plams_init


def model_to_settings(model: str):
    """
    Returns Settings
    """
    settings = Settings()
    model = model.lower()
    if model == "uff":
        settings.input.ForceField.Type = "UFF"
    elif model == "gaff":
        settings.input.ForceField.Type = "GAFF"
        settings.input.ForceField.AnteChamberIntegration = "Yes"
    elif model == "gfnff":
        settings.input.GFNFF
    elif model == "ani-2x" or model == "ani2x":
        settings.input.MLPotential.Model = "ANI-2x"
    elif model == "gfn1xtb" or model == "gfn1-xtb":
        settings.input.DFTB.Model = "GFN1-xTB"
    elif model == "m3gnet-up-2022" or model == "m3gnetup2022":
        settings.input.MLPotential.Model = "M3GNet-UP-2022"
    else:
        raise ValueError("Unknown model: {}".format(model))

    return settings


def _get_quick_settings(model, settings, nproc):
    if settings is None:
        my_settings = model_to_settings(model)
    else:
        my_settings = settings.copy()

    if nproc:
        my_settings.runscript.nproc = nproc

    return my_settings



from collections import OrderedDict
from itertools import product

from scm.plams.core.basejob import MultiJob
from scm.plams.core.results import Results

__all__ = ["NumGradJob", "NumGradResults"]  # names exported to the main namespace


class NumGradResults(Results):

    def _get_gradient_component(self, atom, coord, extract):
        """Get gradient component for *atom* along *coord*.

        *atom* should be integer, *coord* a single character string with 'x', 'y' or 'z'.
        *extract* should be a function that takes results of a single point job and
        returns a single number.
        """
        energies = [extract(self.job.children[(atom, coord, i)].results) for i in self.job.steps]
        coeffs, denom = self.job._coeffs[self.job.npoints]
        return sum([c * e for c, e in zip(coeffs, energies)]) / (denom * self.job.step)

    def get_gradient(self, extract):
        """Get the full gradient vector. Returns a list of length 3N,
        where N is the number of atoms in the calculated molecule.

        *extract* should be a function that takes results of a single point job and
        returns a single number.
        """
        return [
            self._get_gradient_component(atom, coord, extract)
            for atom, coord in product(range(1, 1 + len(self.job.molecule)), "xyz")
        ]


class NumGradJob(MultiJob):
    """A class for calculating numerical gradients of energy
    (or some other single-valued property) with respect to
    cartesian displacements.

    The differentiation is done using the final difference method
    with 2, 4, 6 or 8 points. The length of the step can be adjusted.
    """

    _result_type = NumGradResults

    # finite difference coefficients for different number of points:
    _coeffs = {
        2: ([-1, 1], 2),
        4: ([1, -8, 8, -1], 12),
        6: ([-1, 9, -75, 75, -9, 1], 60),
        8: ([3, -32, 168, -672, 672, -168, 32, -3], 840),
    }

    def __init__(self, molecule, npoints=2, step=0.01, unit="angstrom", jobtype=None, **kwargs):
        # initialize parent class and store all additional constructor arguments
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)
        self.molecule = molecule
        self.npoints = npoints
        self.step = step
        self.unit = unit
        self.jobtype = jobtype  # who is going to calculate single points

    def prerun(self):  # noqa F811
        self.steps = list(range(-(self.npoints // 2), 0)) + list(range(1, self.npoints // 2 + 1))

        for atom, axis, i in product(range(1, 1 + len(self.molecule)), "xyz", self.steps):
            v = (
                self.step * i if axis == "x" else 0,
                self.step * i if axis == "y" else 0,
                self.step * i if axis == "z" else 0,
            )
            newmol = self.molecule.copy()
            newmol[atom].translate(v, self.unit)
            newname = "{}_{}_{}".format(atom, axis, i)
            # settings of NumGradJob are directly passed to children single point jobs
            self.children[(atom, axis, i)] = self.jobtype(name=newname, molecule=newmol, settings=self.settings)



from scm.plams.core.errors import ResultsError
from scm.plams.core.functions import log
from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults
from scm.plams.tools.units import Units

__all__ = ["BANDJob", "BANDResults"]


class BANDResults(SCMResults):
    _kfext = ".runkf"
    _rename_map = {"RUNKF": "$JN" + _kfext}

    def get_properties(self):
        """get_properties()
        Return a dictionary with all the entries from ``Properties`` section in the main KF file (``$JN.rkf``).
        """
        n = self.readkf("Properties", "nEntries")
        ret = {}
        for i in range(1, n + 1):
            tp = self.readkf("Properties", "Type({})".format(i)).strip()
            stp = self.readkf("Properties", "Subtype({})".format(i)).strip()
            val = self.readkf("Properties", "Value({})".format(i))
            if stp.endswith(" (a.u.)"):
                stp = stp[:-7]
            key = "{} {}".format(stp, tp.lower())
            ret[key] = val
        return ret

    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Molecule`` section in the main KF file (``$JN.runkf``).

        For runs with multiple geometries (geometry optimization, transition state search, molecular dynamics) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please use :meth:`get_input_molecule` or extract coordinates from section ``History``, variables ``xyz0``, ``xyz1`` and so on. Mind the fact that all coordinates written by BAND to ``History`` section are in bohr and internal atom order::

            mol = results.get_molecule(section='History', variable='xyz0', unit='bohr', internal=True)
        """
        ret = self.get_molecule(section="Molecule", variable="Coords", unit="bohr", internal=True)
        ret.properties.charge = self.readkf("Molecule", "Charge")
        if ("Molecule", "LatticeVectors") in self._kf:
            lattice = self.readkf("Molecule", "LatticeVectors")
            lattice = Units.convert(lattice, "bohr", "angstrom")
            ret.lattice = [tuple(lattice[i : i + 3]) for i in range(0, len(lattice), 3)]
        return ret

    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.runkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ("History", "nr of geometries") in self._kf:
            ret = self.get_molecule(section="History", variable="xyz0", unit="bohr", internal=False)
            lattice = self.readkf("History", "lattice0")
            lattice = Units.convert(lattice, "bohr", "angstrom")
            lattice = [lattice[i : i + 3] for i in range(0, len(lattice), 3)]
            while len(lattice) > 0 and not any(lattice[-1]):
                lattice.pop()
            ret.lattice = [tuple(i) for i in lattice]
            return ret
        return self.get_main_molecule()

    def get_energy(self, unit="au"):
        """get_energy(unit='au')
        Return final bond energy, expressed in *unit*.
        """
        return self._get_single_value("Bond energies", "final bond energy", unit)

    def get_fermi_energy(self, unit="au"):
        """get_fermi_energy(unit='au')
        Return Fermi energy, expressed in *unit*.
        """
        return self._get_single_value("BandStructure", "FermiEnergy", unit)

    def get_band_gap(self, unit="au"):
        """get_band_gap(unit='au')
        Return the value of the band gap, expressed in *unit*.
        """
        return self._get_single_value("BandStructure", "BandGap", unit)

    def get_dipole_vector(self, unit="au"):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if "Dipole moment vector" in prop:
            return Units.convert(prop["Dipole moment vector"], "au", unit)
        raise ResultsError("'Dipole moment vector' not present in 'Properties' section of {}".format(self._kfpath()))

    def _int2inp(self):
        """_int2inp()
        Get mapping from the internal atom order to the input atom order.
        """
        return self.readkf("geometry", "Atom map new order")

    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        mapping = self._int2inp()
        atnums = self.readkf("Molecule", "AtomicNumbers")
        return [atnums[mapping[i] - 1] for i in range(len(atnums))]


class BANDJob(SCMJob):
    _result_type = BANDResults
    _command = "band"

    def _serialize_mol(self):
        s = self.settings.input
        s.units.length = "angstrom"

        if len(self.molecule.lattice) in [1, 2] and self.molecule.align_lattice():
            log(
                "The lattice supplied for job {} did not follow the convention required by BAND. I rotated the whole system for you. You're welcome".format(
                    self._full_name()
                ),
                3,
            )

        for i, atom in enumerate(self.molecule):
            s.atoms["_" + str(i + 1)] = atom.str(symbol=self._atom_symbol(atom), space=18, decimal=10)

        if self.molecule.lattice:
            for i, vec in enumerate(self.molecule.lattice):
                s.lattice["_" + str(i + 1)] = "{:16.10f} {:16.10f} {:16.10f}".format(*vec)

    def _remove_mol(self):
        if "atoms" in self.settings.input:
            del self.settings.input.atoms
        if "lattice" in self.settings.input:
            del self.settings.input.lattice



from scm.plams.core.errors import FileError, PlamsError
from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults

__all__ = ["AMSAnalysisJob", "AMSAnalysisResults", "convert_to_unicode"]


class AMSAnalysisPlot:
    """
    Class representing a plot of 2D or higher

    * ``x``       -- A list of lists containing the values in each of the multiple x-axes
    * ``y``       -- A list containing the values along the y-axis
    * ``y_sigma`` -- A list containing the standard deviation of the values onthe y-axis
    * ``name``    -- The name of the plot

    The most important method is the write method, which returns a string containing all the plot info,
    and can also write a corresponding file if a filename is provided as argument.
    This file can be read by e.g. gnuplot.
    """

    def __init__(self):
        """
        Initiate an instance of the plot class
        """
        self.x = []
        self.x_units = []
        self.x_names = []

        self.y = None
        self.y_units = None
        self.y_name = None
        self.y_sigma = None  # stadard deviation for y_values

        self.properties = None
        self.name = None
        self.section = None

    def read_data(self, kf, sec):
        """
        Read the xy data for a section from the kf file
        """
        # Read all the x-values. There can be multiple axes for ND plots (n=3,4,....)
        sections = kf.get_skeleton()
        xkeys = [k for k in sections[sec] if "x(" in k and ")-axis" in k]
        xnums = sorted([int(k.split("(")[1].split(")")[0]) for k in xkeys])
        xnums = sorted([xnum for xnum in set(xnums)])
        for i in xnums:
            xkey = "x(%i)-axis" % (i)
            self.x.append(kf.read(sec, xkey))
            x_name = kf.read(sec, "%s(label)" % (xkey))
            self.x_names.append(convert_to_unicode(x_name))
            self.x_units.append(convert_to_unicode(kf.read(sec, "%s(units)" % (xkey))))

        # Read the y-values
        ykey = "y-axis"
        y_name = kf.read(sec, "%s(label)" % (ykey))
        self.y = kf.read(sec, ykey)
        self.y_name = convert_to_unicode(y_name)
        self.y_units = convert_to_unicode(kf.read(sec, "%s(units)" % (ykey)))

        self.y_sigma = kf.read(sec, "sigma")

        self.read_properties(kf, sec)
        self.section = sec.split("(")[0] + "_" + sec.split("(")[1].split(")")[0]
        self.name = self.section

    def read_properties(self, kf, sec):
        """
        Read properties from the KF file
        """
        counter = 0
        properties = {}
        while 1:
            counter += 1
            try:
                propname = kf.read(sec, "Property(%i)" % (counter)).strip()
            except:
                break
            properties[propname] = kf.read(sec, propname)
            if isinstance(properties[propname], str):
                properties[propname] = properties[propname].strip()
                properties[propname] = convert_to_unicode(properties[propname])

        # Now set the instance variables
        self.properties = properties
        if "Legend" in properties:
            self.name = properties["Legend"]

    def get_dimensions(self):
        """
        Get the dimensonality of the plot
        """
        return len(self.x)

    def write(self, outfilename=None):
        """
        Print this plot to a text file
        """
        # Place property string
        parts = []
        for propname, prop in self.properties.items():
            parts.append("%-30s %s\n" % (propname, prop))

        # Place the string with the column names
        x_name = ""
        for xname, xunit in zip(self.x_names, self.x_units):
            x_str = "%s(%s)" % (xname, xunit)
            x_name += "%30s " % (x_str)
        y_name = "%s(%s)" % (self.y_name, self.y_units)
        parts.append("%s %30s %30s\n" % (x_name, y_name, "sigma"))

        # Determine the number of values per axis
        ndims = len(self.x)
        axis_length = int(len(self.x[0]) ** (1.0 / ndims))

        # Place the values
        value_lists = self.x + [self.y] + [self.y_sigma]
        for i, values in enumerate(zip(*value_lists)):
            v_str = ""
            for v in values:
                v_str += "%30.10e " % (v)
            v_str += "\n"
            if (i + 1) % axis_length == 0:
                v_str += "\n"
            parts.append(v_str)
        block = "".join(parts)

        if outfilename is not None:
            outfile = open(outfilename, "w", encoding="utf8")
            outfile.write(block)
            outfile.close()

        return block

    @classmethod
    def from_kf(cls, kf, section, i=1):
        xy = cls()

        # Find the correct section in the KF file
        sections = kf.sections()
        matches = [s for s in sections if s.lower() == section.lower() + "(%i)" % (i)]
        if len(matches) == 0:
            print("Sections: ", list(sections))
            raise PlamsError(
                'AMSAnalysisResults.get_xy(section,i): section must be one of the above. You specified "{}"'.format(
                    section
                )
            )
        sec = matches[0]

        # Get the data
        xy.read_data(kf, sec)
        return xy


class AMSAnalysisResults(SCMResults):
    _kfext = ".kf"
    _rename_map = {"plot.kf": "$JN" + _kfext}

    def get_molecule(self, *args, **kwargs):
        raise PlamsError("AMSAnalysisResults does not support the get_molecule() method.")

    def get_sections(self):
        """
        Read the sections available to make xy plots
        """
        if not self._kfpresent():
            raise FileError("File {} not present in {}".format(self.job.name + self.__class__._kfext, self.job.path))
        if self._kf.reader._sections is None:
            self._kf.reader._create_index()
        sections = self._kf.reader._sections.keys()
        return sections

    def get_xy(self, section="", i=1):
        """
        Get the AMSAnalysitPlot object for a specific section of the plot KFFile
        """
        task = self.job.settings.input.Task
        if section == "":
            section = task

        if not self._kfpresent():
            raise FileError("File {} not present in {}".format(self.job.name + self.__class__._kfext, self.job.path))
        xy = AMSAnalysisPlot.from_kf(self._kf, section, i)
        return xy

    def get_all_plots(self):
        """
        Get a list of all the plot objects created by the analysis jobs
        """
        sections = self.get_sections()
        plots = []
        for section in sections:
            if section == "General":
                continue
            if "History" in section:
                continue
            name_part = section.split("(")[0]
            num_part = int(section.split("(")[1].split(")")[0])
            xy = self.get_xy(name_part, num_part)
            plots.append(xy)
        return plots

    def write_all_plots(self):
        """
        Write all the plots created by the analysis job to file
        """
        plots = self.get_all_plots()
        for xy in plots:
            xy.write("%s" % (xy.section + ".dat"))

    def get_D(self, i=1):
        """returns a 2-tuple (D, D_units) from the AutoCorrelation(i) section on the .kf file."""

        # If there are multiple, it will read the first one
        sections = [sec for sec in self.get_sections() if "Integral" in sec]
        if len(sections) < i:
            return None, None
        section = sections[i - 1]
        plot = self.get_xy(section.split("(")[0], i)
        if not "DiffusionCoefficient" in plot.properties.keys():
            return None, None

        D = plot.properties["DiffusionCoefficient"]
        D_units = plot.y_units
        return D, D_units


class AMSAnalysisJob(SCMJob):
    """A class for analyzing molecular dynamics trajectories using the ``analysis`` program."""

    _result_type = AMSAnalysisResults
    _command = "analysis"
    _subblock_end = "end"

    def __init__(self, **kwargs):
        SCMJob.__init__(self, **kwargs)

    def _serialize_mol(self):
        pass

    def _remove_mol(self):
        pass

    def check(self):
        try:
            grep = self.results.grep_file("$JN.err", "NORMAL TERMINATION")
        except:
            return False
        return len(grep) > 0


def convert_to_unicode(k):
    """
    Convert a string with ascii symbols representing unicode symbols

    Example k: 'abc\\u03c9def'
    """
    parts = k.split("\\u")
    # Collect the hexadecimals
    symbols = [chr(int(part[:4], 16)) for part in parts[1:]]
    # Now repair the parts
    parts = parts[:1] + ["".join([s, part[4:]]) for s, part in zip(symbols, parts[1:])]
    key = "".join(parts)

    return key



import copy
import hashlib
import os
import subprocess
import sys
import time
import warnings
from contextlib import AbstractContextManager
from os.path import join as opj
from typing import Callable, Dict, NoReturn

__all__ = []


def smart_copy(obj, owncopy=[], without=[]):
    """Return a copy of *obj*. Attributes of *obj* listed in *without* are ignored. Attributes listed in *owncopy* are copied by calling their own ``copy()`` methods. All other attributes are copied using :func:`copy.deepcopy`."""

    ret = obj.__class__()
    for k in owncopy:
        ret.__dict__[k] = obj.__dict__[k].copy()
    for k in obj.__dict__:
        if k not in (without + owncopy):
            ret.__dict__[k] = copy.deepcopy(obj.__dict__[k])
    return ret


# ===========================================================================


def sha256(string):
    """A small utility wrapper around :ref:`hashlib.sha256<hash-algorithms>`."""
    if not isinstance(string, bytes):
        string = str(string).encode()
    h = hashlib.sha256()
    h.update(string)
    return h.hexdigest()


# ===========================================================================


def saferun(*args, **kwargs):
    """A wrapper around :func:`subprocess.run` repeating the call ``config.saferun.repeat`` times with ``config.saferun.delay`` interval in case of :exc:`BlockingIOError` being raised (any other exception is not caught and directly passed above). All arguments (*args* and *kwargs*) are passed directly to :func:`~subprocess.run`. If all attempts fail, the last raised :exc:`BlockingIOError` is reraised."""
    from scm.plams.core.functions import config, log

    attempt = 0
    (repeat, delay) = (config.saferun.repeat, config.saferun.delay) if ("saferun" in config) else (5, 1)
    while attempt <= repeat:
        try:
            return subprocess.run(*args, **kwargs)
        except OSError as e:
            attempt += 1
            log("subprocess.run({}) attempt {} failed with {}".format(args[0], attempt, e), 5)
            last_error = e
            time.sleep(delay)
    raise last_error


# ===========================================================================


class UpdateSysPath(AbstractContextManager):
    """A context manager for temporary adding ``$AMSHOME/scripting`` to :data:`sys.path`.

    While the context manager is opened modules can be imported (directly) from the ``$AMSHOME/scripting`` directory,
    allowing one access to the python modules distributed with ADF.
    An :exc:`EnvironmentError` is raised if the ``AMSHOME`` environment variable has not been set.

    Usage example:

    .. code:: python
        >>> import sys
        >>> import os

        >>> scripting = os.path.join(os.environ['AMSHOME'], 'scripting')
        >>> with UpdateSysPath():
        ...     print(scripting in sys.path)  # The context manager is opened
        True

        >>> print(scripting in sys.path)  # The context manager is closed
        False

    """

    def __init__(self, path: str = None) -> None:
        try:
            parser_path = path if path is not None else opj(os.environ["AMSHOME"], "scripting")
        except KeyError as ex:
            raise EnvironmentError("The 'AMSHOME' environment variable has not been set").with_traceback(
                ex.__traceback__
            )

        if parser_path not in sys.path:
            sys.path.append(parser_path)
            self.path = parser_path
        else:  # The specified path is already present in sys.path
            self.path = None

    def __enter__(self) -> None:
        """If not ``None``, add :attr:`UpdateSyspath.path` to :data:`sys.path`."""
        if self.path is not None:
            sys.path.append(self.path)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """If not ``None``, remove :attr:`UpdateSyspath.path` from :data:`sys.path`."""
        if self.path is not None:
            try:
                idx = sys.path.index(self.path)
            except ValueError:  # self.path has been (manually) removed by the user
                pass
            else:
                del sys.path[idx]


# ===========================================================================


def _raise(exc: Exception) -> NoReturn:
    """Raise *exc*; this tiny function is here because ``lambda`` expressions can't handle the ``raise`` statement."""
    raise exc


#: Map a :class:`str` representation of an action to an actual callable.
_ACTION_DICT: Dict[str, Callable[[Exception], None]] = {
    "ignore": lambda n: None,
    "raise": lambda n: _raise(n),
    "warn": lambda n: warnings.warn(str(n), stacklevel=2),
}


def parse_action(action: str) -> Callable[[Exception], None]:
    """Return a callable which takes an :exc:`Exception` as argument and performs some action base on the value of *action*.

    * ``action = 'ignore'``: Do nothing.
    * ``action = 'warn'``: Issue a warning.
    * ``action = 'raise'``: Raise the passed exception.

    """
    try:
        return _ACTION_DICT[action]
    except KeyError as ex:
        raise ValueError("`action` expected 'ignore', 'raise' or 'warn'; " f"observed value: {action!r}") from ex


# ===========================================================================


def retry(sleep=0.1, maxtries=10):
    # wrapper for sleep-retrying a function call. use with `@retry()`
    from time import sleep as _sleep

    def wrap1(f):
        def wrap2(*a, __count=0, **kw):
            try:
                return f(*a, **kw)
            except Exception as e:
                if __count > maxtries:
                    raise e from None  # ignores stack trace
                _sleep(sleep)
                wrap2(
                    *a,
                    __count=__count + 1,
                    **kw,
                )

        return wrap2

    return wrap1



#!/usr/bin/env plams
import multiprocessing

# Run jobs as many jobs as possible in parallel:
config.default_jobrunner = JobRunner(parallel=True, maxjobs=multiprocessing.cpu_count())
config.job.runscript.nproc = 1

# The molecules we want to use in our benchmark:
mol_smiles = {"Methane": "C", "Ethane": "C-C", "Ethylene": "C=C", "Acetylene": "C#C"}
molecules = {}
for name, smiles in mol_smiles.items():
    # Compute 10 conformers, optimize with UFF and pick the lowest in energy.
    molecules[name] = from_smiles(smiles, nconfs=10, forcefield="uff")[0]
    print(name, molecules[name])

# Initialize the common settings:
common_settings = Settings()
common_settings.input.ams.Task = "SinglePoint"
common_settings.input.ams.System.Symmetrize = "Yes"
common_settings.input.adf.Basis.Core = "None"

basis = ["QZ4P", "TZ2P", "TZP", "DZP", "DZ", "SZ"]
reference_basis = "QZ4P"

# Set up and run the calculations:
results = {}
for bas in basis:
    for name, molecule in molecules.items():
        settings = common_settings.copy()
        settings.input.adf.Basis.Type = bas
        job = AMSJob(name=name + "_" + bas, molecule=molecule, settings=settings)
        results[(name, bas)] = job.run()

# Calculate the average absolute error in bond energy per atom for each basis set:
average_errors = {}
for bas in basis:
    if bas != reference_basis:
        errors = []
        for name, molecule in molecules.items():
            reference_energy = results[(name, reference_basis)].get_energy(unit="kcal/mol")
            energy = results[(name, bas)].get_energy(unit="kcal/mol")
            errors.append(abs(energy - reference_energy) / len(molecule))
            print("Energy for {} using {} basis set: {} [kcal/mol]".format(name, bas, energy))
        average_errors[bas] = sum(errors) / len(errors)

print("== Results ==")
print("Average absolute error in bond energy per atom")
for bas in basis:
    if bas != reference_basis:
        print("Error for basis set {:<4}: {:>10.3f} [kcal/mol]".format(bas, average_errors[bas]))



"""
Run RASPA2 with PLAMS
Contributed by Patrick Melix

See the Documentation for an Example
"""

import shutil
from os import symlink
from os.path import join as opj

from scm.plams.core.basejob import SingleJob
from scm.plams.core.errors import JobError, ResultsError
from scm.plams.core.results import Results
from scm.plams.core.settings import Settings

__all__ = ["RaspaJob", "RaspaResults"]


class RaspaResults(Results):
    """A Class for handling RASPA Results.

    Note that the Output file is guessed in ``self.collect``. Do not rely on it to point to the
    file you need and but check ``self._filenames['out']``!
    """

    def collect(self):
        """Try to Guess the Main Output File.

        Set it to whatever comes first in the 'Output/System_0' folder.
        """
        Results.collect(self)
        if "out" not in self.job._filenames:
            files = [f for f in self.files if "Output/System_0" in f]
            if files:
                self.job._filenames["out"] = files[0]

    def get_block_value(self, search, file=None, get_std=False, get_unit=False):
        """Return a Block Value from Output File.

        Can be used to retrieve block values from the final section of the output files.
        Uses `Results.grep_file` to retrieve the values.
        Remember to escape special `grep` characters in the search string.
        Returns the average value over all blocks.

        search: string
        String labeling the value you desire.

        file: string
        Defaults to the automatically detected output file. See `self.collect()`.

        get_std: boolean
        Also return the standard deviation of the value

        get_unit: boolean
        Also return the unit of the value
        """
        if not file:
            file = self.job._filenames["out"]

        chunk = self.grep_file(file, pattern=search, options="-A 8")
        last_line = chunk[-1].strip()
        if not last_line.startswith("Average"):
            raise ResultsError("Unable to get the requested value from {:}".format(file))

        last_line = last_line.split()
        ret = [float(last_line[1])]

        if get_std:
            ret.append(float(last_line[4]))
        if get_unit:
            ret.append(last_line[-1][1:-1])

        if len(ret) > 1:
            return tuple(ret)
        else:
            return ret[0]

    def get_value(self, search, file=None, get_std=False, get_unit=False):
        """Return a Single Value from Output File.

        Can be used to retrieve single values from the final section of the output files.
        Uses `Results.grep_file` to retrieve the values.
        Remember to escape special `grep` characters in the search string.

        search: string
        String labeling the value you desire.

        file: string
        Defaults to the automatically detected output file. See `self.collect()`.

        get_std: boolean
        Also return the standard deviation of the value

        get_unit: boolean
        Also return the unit of the value
        """
        if not file:
            file = self.job._filenames["out"]

        line = self.grep_file(file, pattern=search)[-1].split()
        # Two cases, with standard deviation and without
        if "+/-" in line:
            idx = line.index("+/-") - 1
            ret = [float(line[idx])]
            if get_std:
                ret.append(float(line[idx + 2]))
        else:
            if get_std:
                raise ResultsError("Did not find +/- in {:}. Does the value you requested have one?!".format(file))
            # extract first float
            for s in line:
                try:
                    ret = [float(s)]
                    break
                except BaseException:
                    continue
        # unit always in [], sometimes at the end, sometimes in the middle.
        # always use the first [] to find the unit (sometimes it is given twice
        # and the second one is empty)
        if get_unit:
            line = " ".join(line)
            idx1 = line.index("[")
            idx2 = line.index("]")
            ret.append(line[idx1 + 1 : idx2])

        if len(ret) > 1:
            return tuple(ret)
        else:
            return ret[0]

    def get_from_all_files(self, *args, output_folder="Output/System_0/", method="get_value", **kwargs):
        """Wrapper to Execute Methods on all Output files.

        output_folder: string
        The subfolder containing the relevant output files. Default should be fine.

        method: function
        The function to call, needs to be a function of the |RaspaResults| class that takes a *file* argument.
        All *args* and *kwargs* are passed on.

        If the *method* returns tuples, the tuples are unpacked and added to lists.
        """
        call_method = getattr(self, method)
        relevant_files = [f for f in self.files if output_folder in f]
        ret = []
        for f in relevant_files:
            data = call_method(*args, file=f, **kwargs)
            if isinstance(data, tuple):
                # initialize ret list to proper length
                if len(ret) == 0:
                    ret = [[] for i in range(len(data))]
                for i, d in enumerate(data):
                    ret[i].append(d)
            else:
                ret.append(data)

        if isinstance(ret[0], list):
            return tuple(ret)
        else:
            return ret

    def get_isotherm(
        self,
        output_folder="Output/System_0/",
        search_x="Partial pressure:",
        search_y=r"Average loading excess \[cm^3 (STP)/gr",
        get_std=False,
        get_unit=False,
    ):
        """Try to automatically collect Isotherm Data.

        output_folder: string
        The subfolder containing the relevant output files. Default should be fine.

        search_x: string
        String labeling the desired x-Axis values in the output.

        search_y: string
        String labeling the desired y-Axis values in the output.

        get_std: boolean
        Return the standard deviation of the y values as a list.

        get_unit: boolean
        Return the units of the values as strings `x_unit`, `y_unit`.

        Returns two lists: `x_values`, `y_values`. Optionally after that `x_std`, `x_unit`, `y_unit`.
        """
        ret = []
        x_info = self.get_from_all_files(search_x, output_folder=output_folder, get_unit=get_unit)
        y_info = self.get_from_all_files(search_y, output_folder=output_folder, get_std=get_std, get_unit=get_unit)

        if not get_std and not get_unit:
            ret = [x_info, y_info]
        elif get_std and not get_unit:
            ret = [x_info, *y_info]
        elif get_std:
            ret = [x_info[0], *y_info[0:2], x_info[1][0], y_info[2][0]]  # only one unit
        else:
            ret = [x_info[0], y_info[0], x_info[1][0], y_info[1][0]]  # only one unit

        return tuple(ret)


class RaspaJob(SingleJob):
    """A single job with RASPA2 class.

    Requires the ASE Python package.

    Files that should be copied to the job directory can be passed as a dictionary using the ``copy``
    argument. The dict should have the following layout: *dict[<filename in jobdir>] = <path to file>*.
    If you prefer symlinks instead of copies use the ``symlink`` argument with the same layout.
    The path strings given to symlink are not altered. Remember that the job directory is not equal
    to the current work directory if using relative symlinks.

    **Molecule parsing is not yet supported!**

    The environment variable ``$RASPA_DIR`` needs to be set for the runscript to work.
    """

    _result_type = RaspaResults
    _filenames = {"inp": "simulation.input", "run": "$JN.run", "err": "$JN.err", "stdout": "$JN.out"}

    def __init__(self, copy=None, symlink=None, **kwargs):
        SingleJob.__init__(self, **kwargs)
        self.copy_files = copy
        self.symlink_files = symlink
        # Output Name is not known before running
        self.settings.runscript.stdout_redirect = True

    def _get_ready(self):
        """Copy files to execution dir if self.copy_files is set."""
        SingleJob._get_ready(self)
        if self.copy_files:
            for filename, path in self.copy_files.items():
                dest = opj(self.path, filename)
                shutil.copy(path, dest)
        if self.symlink_files:
            for filename, path in self.symlink_files.items():
                dest = opj(self.path, filename)
                symlink(path, dest)
        return

    def get_input(self):
        """Transform all contents of ``settings.input`` branch into string with blocks, keys and values.

        RASPA2 does not interpret indentation but rather certain keys automatically start a subsection.
        Subsections are never closed but *automatically* ended if the next key-value pair does not fit into it.
        So here we only need to care about readability of the input file. The logic is automatically preserved
        by using the nested PLAMS ``settings`` logic.

        Components should be built using either ``component['INT'] = Settings()`` or ``component._INT = Settings()``.
        It needs to have a key ``moleculename`` (case insensitive).
        """

        def parse(key, value, indent=""):
            # transform bools to yes and no
            if value is True:
                value = "yes"
            elif value is False:
                value = "no"

            ret = indent + key

            if key.lower() == "component":
                n_comp = len(value)
                for i in range(n_comp):
                    # get component number
                    if str(i) in value:
                        number = str(i)
                    elif "_{}".format(i) in value:
                        number = "_{}".format(i)
                    else:
                        raise JobError(
                            "Check that you have a valid component {:} in settings.input.component".format(i)
                        )
                    ret += " " + str(i)

                    subkeys = list(value[number].keys())
                    mol_name = subkeys[[sk.lower() for sk in subkeys].index("moleculename")]
                    ret += " MoleculeName " + str(value[number][mol_name])
                    ret += "\n"
                    tmp_value = value[number].copy()
                    del tmp_value[mol_name]  # exclude MoleculeName
                    for el in tmp_value:
                        ret += parse(el, tmp_value[el], indent + "  ")
                    ret += "\n" + indent + key

                return ret

            elif isinstance(value, Settings):
                if "_h" in value:
                    ret += " " + str(value["_h"])
                ret += "\n"

                for el in value:
                    if el != "_h":
                        ret += parse(el, value[el], indent + "  ")
            else:
                ret += " " + str(value)
            ret += "\n"
            return ret

        inp = ""
        for item in self.settings.input:
            inp += parse(item, self.settings.input[item])

        return inp

    def get_runscript(self):
        """$RASPA_DIR has to be set in the environment!"""
        ret = "export DYLD_LIBRARY_PATH=${RASPA_DIR}/lib"
        ret += "\nexport LD_LIBRARY_PATH=${RASPA_DIR}/lib"
        ret += "\n$RASPA_DIR/bin/simulate "
        if self.settings.runscript.stdout_redirect:
            ret += " >" + self._filename("stdout")
        ret += "\n\n"
        return ret

    def check(self):
        """Look for the normal termination signal in output."""
        s = self.results.grep_output("Simulation finished on")
        return bool(s)



#!/usr/bin/env amspython
from scm.plams import *


def main():
    init()

    mol = from_smiles("O")
    mol.lattice = [
        [
            3.0,
            0.0,
            0.0,
        ],
        [0.0, 3.0, 0.0],
        [0.0, 0.0, 3.0],
    ]

    s = Settings()
    s.runscript.nproc = 1
    s.input.ams.task = "GeometryOptimization"
    s.input.ams.GeometryOptimization.Convergence.Gradients = 0.01  # hartree/ang

    s.input.MLPotential.Model = "M3GNet-UP-2022"
    # If you have trained a custom M3GNet model yourself, you can use:
    # s.input.MLPotential.Model = 'Custom'
    # s.input.MLPotential.Backend = 'm3gnet'
    # s.input.MLPotential.ParameterDir = '<path to directory containing the M3GNet model>'

    job = AMSJob(settings=s, molecule=mol, name="ams_with_m3gnet")
    job.run()

    energy = job.results.get_energy(unit="eV")
    print(f"M3GNet: final energy {energy:.3f} eV")

    finish()


if __name__ == "__main__":
    main()



#!/usr/bin/env plams
# Compute the neutral-anion reorganization energy of pyrrole
# using ADF as computational engine

molecule = Molecule("pyrrole.xyz")

# Generic settings of the calculation
# (for quantitatively better results, use better settings)
common_settings = Settings()
common_settings.input.adf.Basis.Type = "DZ"

# Specific settings for the neutral calculation.
# Nothing special needs to be done for the neutral calculation,
# so we just use an empty settings.
neutral_settings = Settings()

# Specific settings for the anion calculation:
anion_settings = Settings()
anion_settings.input.ams.System.Charge = -1
anion_settings.input.adf.Unrestricted = "Yes"
anion_settings.input.adf.SpinPolarization = 1

# Create and run the ReorganizationEnergyJob:
job = ReorganizationEnergyJob(
    molecule, common_settings, neutral_settings, anion_settings, name=molecule.properties.name
)
job.run()

# Fetch and print the results:

energy_unit = "eV"
energies = job.results.get_all_energies(energy_unit)
reorganization_energy = job.results.reorganization_energy(energy_unit)

print("")
print("== Results ==")
print("")
print(f"Molecule: {molecule.properties.name}")
print("State A: neutral")
print("State B: anion")
print("")
print(f"Reorganization energy: {reorganization_energy:.6f} [{energy_unit}]")
print("")
print(f"|   State   | Optim Geo | Energy [{energy_unit}]")
print(f'|     A     |     A     | {energies["state A geo A"]:.6f}')
print(f'|     A     |     B     | {energies["state A geo B"]:.6f}')
print(f'|     B     |     A     | {energies["state B geo A"]:.6f}')
print(f'|     B     |     B     | {energies["state B geo B"]:.6f}')
print("")



import os

from scm.plams.core.functions import log
from scm.plams.interfaces.adfsuite.ams import AMSJob

__all__ = ["ADFNBOJob"]


class ADFNBOJob(AMSJob):

    def prerun(self):  # noqa F811
        s = self.settings.input.ADF
        s.fullfock = True
        s.aomat2file = True
        s.symmetry = "NoSym"
        s.basis.core = "None"
        if "save" in s:
            if isinstance(s.save, str):
                s.save += " TAPE15"
            elif isinstance(s.save, list):
                s.save.append("TAPE15")
            else:
                log(
                    "WARNING: 'SAVE TAPE15' could not be added to the input settings of {}. Make sure (thisjob).settings.input.save is a string or a list.".format(
                        self.name
                    ),
                    1,
                )
        else:
            s.save = "TAPE15"

        if isinstance(self.settings.adfnbo, list):
            adfnbo_input = self.settings.adfnbo
        else:
            adfnbo_input = ["write", "spherical", "fock"]
            log(
                "WARNING: (thisjob).settings.adfnbo should be a list. Using default settings: write, fock, spherical", 1
            )

        self.settings.runscript.post = (
            'cp "' + os.path.join(self.path, "adf.rkf") + '" TAPE21\n'
            "$AMSBIN/adfnbo <<eor\n" + "\n".join(adfnbo_input) + "\neor\n\n$AMSBIN/gennbo6 FILE47\n"
        )



#!/usr/bin/env plams
import matplotlib.pyplot as plt

"""
Example showing the UseLowestEnergy feature of the Hybrid engine and the OptimizeSpinRound feature of the ADF engine.

A bond scan is performed on hydrogen peroxide with UFF. The resulting
structures are then recalculated with DFT using four different setups. In
particular, one setup uses the Hybrid engine with DynamicFactors=UseLowestEnergy.
Another uses the ADF engine with OptimizeSpinRound in combination with an ElectronicTemperature.

In the end, the results of energy vs. bond length are plotted.

To run this example:
$AMSBIN/plams UseLowestEnergy.py
"""


def main():
    # config.job.runscript.nproc = 1 # uncomment to run all jobs in serial

    pesscan_job = initial_pesscan()
    rkf = pesscan_job.results.rkfpath()
    singlet_job = replay_job(rkf, "singlet")
    triplet_job = replay_job(rkf, "triplet")
    hybrid_job = replay_job(rkf, "hybrid")
    optimizespin_job = replay_job(rkf, "optimizespin")

    # or load the finished jobs from disk:
    # pesscan_job = AMSJob.load_external('plams_workdir.002/initial_pesscan/ams.rkf')
    # rkf = pesscan_job.results.rkfpath()
    # singlet_job = AMSJob.load_external('plams_workdir.002/singlet/ams.rkf')
    # triplet_job = AMSJob.load_external('plams_workdir.002/triplet/ams.rkf')
    # hybrid_job = AMSJob.load_external('plams_workdir.002/hybrid/ams.rkf')
    # optimizespin_job = AMSJob.load_external('plams_workdir.002/optimizespin/ams.rkf')

    plot_results(singlet_job, triplet_job, hybrid_job, optimizespin_job)


def initial_pesscan():
    """Run a bond scan for hydrogen peroxide with UFF. Returns the finished job"""
    mol = from_smiles("OO")  # hydrogen peroxide
    s = Settings()
    s.input.ams.Task = "PESScan"
    s.input.ams.PESScan.ScanCoordinate.nPoints = 7
    # Scan O-O bond length (atoms 1 and 2) between 1.2 and 2.5 Å
    s.input.ams.PESScan.ScanCoordinate.Distance = "2 1 1.2 2.5"
    s.input.ForceField.Type = "UFF"
    job = AMSJob(settings=s, molecule=mol, name="initial_pesscan")
    job.run()
    return job


def singlet_settings(header=""):
    s = Settings()
    s.Unrestricted = "No"
    s.XC.GGA = "PBE"
    s.Basis.Type = "DZP"
    s.SCF.Iterations = 100
    s._h = header
    return s


def triplet_settings(header=""):
    s = Settings()
    s.Unrestricted = "Yes"
    s.SpinPolarization = 2
    s.XC.GGA = "PBE"
    s.Basis.Type = "DZP"
    s.SCF.Iterations = 100
    s._h = header
    return s


def hybrid_settings():
    """Look at the file plams_workdir/hybrid/hybrid.in to see how the below is translated into text input for the hybrid engine"""
    s = Settings()
    s.input.Hybrid.Energy.DynamicFactors = "UseLowestEnergy"

    s.input.Hybrid.Energy.Term = ["Region=* EngineID=Singlet", "Region=* EngineID=Triplet"]

    s.input.Hybrid.Engine = [singlet_settings("ADF Singlet"), triplet_settings("ADF Triplet")]
    return s


def replay_job(rkf, engine="hybrid"):
    """Replay the structures from the UFF pesscan with ADF. Use three different engines:
    Hybrid: This will run all structures with both Singlet and Triplet and pick the lowest energy
    Singlet: This runs all structures using the Singlet settings
    Triplet: This runs all structures using the Triplet settings

    Returns the finished job
    """
    s = Settings()
    if engine == "hybrid":
        s = hybrid_settings()
    elif engine == "singlet":
        s.input.adf = singlet_settings()
    elif engine == "triplet":
        s.input.adf = triplet_settings()
    elif engine == "optimizespin":
        s.input.adf = triplet_settings()
        s.input.adf.Occupations = "ElectronicTemperature=300 OptimizeSpinRound=0.05"

    s.input.ams.Task = "Replay"
    s.input.ams.Replay.File = rkf

    job = AMSJob(settings=s, name=engine)
    job.run()
    return job


def plot_results(singlet_job, triplet_job, hybrid_job, optimizespin_job):
    """
    Generate a plot of the energy vs. bond length for the three different jobs.  Saves a plot to pesplot.png.
    """
    bondlengths = singlet_job.results.get_pesscan_results()["RaveledPESCoords"][0]
    bondlengths = Units.convert(bondlengths, "bohr", "angstrom")

    singlet_pes = singlet_job.results.get_pesscan_results()["PES"]
    triplet_pes = triplet_job.results.get_pesscan_results()["PES"]
    hybrid_pes = hybrid_job.results.get_pesscan_results()["PES"]
    hybrid_pes = [x - 0.005 for x in hybrid_pes]  # slightly downshift for visual clarity when plotting
    optimizespin_pes = optimizespin_job.results.get_pesscan_results()["PES"]
    optimizespin_pes = [x - 0.010 for x in optimizespin_pes]  # slightly downshift for visual clarity when plotting

    plt.plot(bondlengths, singlet_pes)
    plt.plot(bondlengths, triplet_pes)
    plt.plot(bondlengths, hybrid_pes)
    plt.plot(bondlengths, optimizespin_pes)
    plt.title("PES Scan for hydrogen peroxide")
    plt.xlabel("O-O bond length (Å)")
    plt.ylabel("Energy (hartree)")
    plt.legend(["Singlet", "Triplet", "Hybrid: Lowest energy (slightly shifted)", "Optimized spin (slightly shifted)"])
    plt.savefig("pesplot.png")
    plt.show()


if __name__ == "__main__":
    main()



from scm.plams.core.functions import add_to_instance
from scm.plams.core.settings import Settings
from scm.plams.interfaces.adfsuite.ams import AMSJob, AMSResults
import numpy as np
from scm.plams.tools.kftools import KFFile
from scm.plams.tools.units import Units
from typing import Union
import scm.plams as plams

__all__ = ["AMSMDJob", "AMSNVEJob", "AMSNVTJob", "AMSNPTJob"]


class AMSMDJob(AMSJob):
    """

    molecule: Molecule
        The initial structure

    name: str
        The name of the job

    settings: Settings
        Settings for the job. You should normally not populate neither settings.input.ams.MolecularDynamics nor settings.input.ams.Task

    velocities: float or AMSJob or str (path/to/ams.rkf) or 2-tuple (path/to/ams.rkf, frame-number)
        If float, it is taken as the temperature. If AMSJob or str, the velocities are taken from the EndVelocities section of the corresponding ams.rkf file. If 2-tuple, the first item must be a path to an ams.rkf, and the second item an integer specifying the frame number - the velocities are then read from History%Velocities(framenumber).

    timestep: float
        Timestep

    samplingfreq: int
        Sampling frequency

    nsteps: int
        Length of simulation

    **Trajectory options**:

    checkpointfrequency: int
        How frequently to write MDStep*.rkf checkpoint files

    writevelocities : bool
        Whether to save velocities to ams.rkf (needed for example to restart from individual frames or to calculate velocity autocorrelation functions)

    writebonds: bool
        Whether to save bonds to ams.rkf

    writemolecules: bool
        Whether to write molecules to ams.rkf

    writeenginegradients: bool
        Whether to save engine gradients (negative of atomic forces) to ams.rkf

    **Thermostat (NVT, NPT) options**:

    thermostat: str
        'Berendsen' or 'NHC'

    tau: float
        Thermostat time constant (fs)

    temperature: float or tuple of floats
        Temperature (K). If a tuple/list of floats, the Thermostat.Duration option will be set to evenly divided intervals.

    thermostat_region: str
        Region for which to apply the thermostat

    **Barostat (NPT) options**:

    barostat: str
        'Berendsen' or 'MTK'

    barostat_tau: float
        Barostat time constant (fs)

    pressure: float
        Barostat pressure (pascal)

    equal: str
        'XYZ' etc.

    scale: str
        'XYZ' etc.

    constantvolume: bool
        Constant volume?

    **Other options**:

    calcpressure: bool
        Whether to calculate pressure for each frame.

    binlog_time: bool
        Whether to log the time at every timestep in the BinLog section on ams.rkf

    binlog_pressuretensor: bool
        Whether to log the pressure tensor at every timestep in the BinLog section on ams.rkf



    """

    default_nsteps = 1000
    default_timestep = 0.25
    default_samplingfreq = 100

    default_thermostat = "NHC"
    default_temperature = 300
    default_tau_multiplier = 400  # get tau by multiplying the timestep with this number
    default_thermostat_region = None

    default_barostat = "MTK"
    default_pressure = 1e5
    default_barostat_tau_multiplier = 4000  # get barostat_tau by multiplying the timestep with this number
    default_scale = "XYZ"
    default_equal = "None"
    default_constantvolume = "False"

    default_checkpointfrequency = 1000000
    default_writevelocities = "True"
    default_writebonds = "True"
    default_writecharges = "True"
    default_writemolecules = "True"
    default_writeenginegradients = "False"

    default_calcpressure = "False"
    default_binlog_time = "False"
    default_binlog_pressuretensor = "False"
    default_binlog_dipolemoment = "False"

    def __init__(
        self,
        velocities=None,
        timestep=None,
        samplingfreq: int = None,
        nsteps: int = None,
        checkpointfrequency=None,
        writevelocities=None,
        writebonds=None,
        writemolecules=None,
        writecharges=None,
        writeenginegradients=None,
        calcpressure=None,
        molecule: Union[plams.Molecule, AMSJob, AMSResults] = None,
        temperature: float = None,
        thermostat=None,
        tau=None,
        thermostat_region=None,
        pressure=None,
        barostat=None,
        barostat_tau=None,
        scale=None,
        equal=None,
        constantvolume=None,
        binlog_time=None,
        binlog_pressuretensor=None,
        binlog_dipolemoment=None,
        _enforce_thermostat=False,
        _enforce_barostat=False,
        **kwargs,
    ):
        """ """

        if isinstance(molecule, AMSJob):
            molecule = molecule.results.get_main_molecule()
        if isinstance(molecule, AMSResults):
            molecule = molecule.get_main_molecule()
        AMSJob.__init__(self, molecule=molecule, **kwargs)

        self.settings.input.ams.Task = "MolecularDynamics"
        mdsett = self.settings.input.ams.MolecularDynamics

        mdsett.TimeStep = timestep or mdsett.TimeStep or self.default_timestep
        mdsett.Trajectory.SamplingFreq = samplingfreq or mdsett.Trajectory.SamplingFreq or self.default_samplingfreq
        mdsett.NSteps = nsteps or mdsett.NSteps or self.default_nsteps

        mdsett.Trajectory.WriteVelocities = (
            str(writevelocities)
            if writevelocities is not None
            else mdsett.Trajectory.WriteVelocities or self.default_writevelocities
        )
        mdsett.Trajectory.WriteBonds = (
            str(writebonds) if writebonds is not None else mdsett.Trajectory.WriteBonds or self.default_writebonds
        )
        mdsett.Trajectory.WriteMolecules = (
            str(writemolecules)
            if writemolecules is not None
            else mdsett.Trajectory.WriteMolecules or self.default_writemolecules
        )
        mdsett.Trajectory.WriteCharges = (
            str(writecharges)
            if writecharges is not None
            else mdsett.Trajectory.WriteCharges or self.default_writecharges
        )
        mdsett.Trajectory.WriteEngineGradients = (
            str(writeenginegradients)
            if writeenginegradients is not None
            else mdsett.Trajectory.WriteEngineGradients or self.default_writeenginegradients
        )
        mdsett.CalcPressure = (
            str(calcpressure) if calcpressure is not None else mdsett.CalcPressure or self.default_calcpressure
        )
        mdsett.Checkpoint.Frequency = (
            checkpointfrequency or mdsett.Checkpoint.Frequency or self.default_checkpointfrequency
        )

        mdsett.BinLog.Time = (
            str(binlog_time) if binlog_time is not None else mdsett.BinLog.Time or self.default_binlog_time
        )
        mdsett.BinLog.PressureTensor = (
            str(binlog_pressuretensor)
            if binlog_pressuretensor is not None
            else mdsett.BinLog.PressureTensor or self.default_binlog_pressuretensor
        )
        mdsett.BinLog.DipoleMoment = (
            str(binlog_dipolemoment)
            if binlog_dipolemoment is not None
            else mdsett.BinLog.DipoleMoment or self.default_binlog_dipolemoment
        )

        if velocities is None and temperature is not None:
            velocities = self._get_initial_temperature(temperature)
        self.settings += self._velocities2settings(velocities)

        if temperature or thermostat or _enforce_thermostat:
            self.settings.update(
                self._get_thermostat_settings(
                    thermostat=thermostat,
                    temperature=temperature,
                    tau=tau,
                    thermostat_region=thermostat_region,
                    nsteps=int(mdsett.NSteps),
                )
            )

        if pressure or barostat or _enforce_barostat:
            self.settings.update(
                self._get_barostat_settings(
                    pressure=pressure,
                    barostat=barostat,
                    barostat_tau=barostat_tau,
                    scale=scale,
                    equal=equal,
                    constantvolume=constantvolume,
                )
            )

    def remove_blocks(self, blocks=None):
        if blocks:
            for block in blocks:
                if block in self.settings.input.ams.MolecularDynamics:
                    del self.settings.input.ams.MolecularDynamics[block]

    @staticmethod
    def _get_initial_temperature(temperature):
        try:
            return temperature[0]
        except TypeError:
            return temperature

    @staticmethod
    def _velocities2settings(velocities):
        s = Settings()
        if isinstance(velocities, int) or isinstance(velocities, float) or velocities is None or velocities is False:
            s.input.ams.MolecularDynamics.InitialVelocities.Type = "Random"
            s.input.ams.MolecularDynamics.InitialVelocities.Temperature = velocities or AMSMDJob.default_temperature
        elif isinstance(velocities, tuple):
            # file and frame number
            f = velocities[0]
            frame = velocities[1]
            vels = KFFile(f).read("MDHistory", f"Velocities({frame})")
            vels = np.array(vels).reshape(-1, 3) * Units.convert(1.0, "bohr", "angstrom")  # angstrom/fs
            s.input.ams.MolecularDynamics.InitialVelocities.Type = "Input"
            values = ""
            for x in vels:
                values += 6 * " " + " ".join(str(y) for y in x) + "\n"
            s.input.ams.MolecularDynamics.InitialVelocities.Values._h = "   # From {} frame {}".format(f, frame)
            s.input.ams.MolecularDynamics.InitialVelocities.Values._1 = values
        else:
            s.input.ams.MolecularDynamics.InitialVelocities.Type = "FromFile"
            s.input.ams.MolecularDynamics.InitialVelocities.File = velocities
        return s

    def get_velocities_from(self, other_job, frame=None, update_molecule=True):
        """
        Function to update the InitialVelocities block in self. It is normally not needed, instead use the e.g. AMSNVEJob.restart_from() function.

        This function can be called in prerun() methods for MultiJobs
        """
        _, velocities, molecule, _ = self._get_restart_job_velocities_molecule(other_job, frame=frame)
        del self.settings.input.ams.MolecularDynamics.InitialVelocities
        self.settings += self._velocities2settings(velocities)

        if update_molecule:
            self.molecule = molecule

    @staticmethod
    def _get_restart_job_velocities_molecule(other_job, frame=None, settings=None, get_velocities_molecule=True):
        """
        other_job: str or some AMSMdJob

        get_velocities_molecule : bool
            Whether to get the velocities and molecule right now. Set to False to not access other_job.results (for use with MultiJob prerun() methods)

        Returns: (other_job [AMSJob], velocities, molecule, extra_settings [Settings])
        """
        if isinstance(other_job, str):
            other_job = AMSJob.load_external(other_job)
        if get_velocities_molecule:
            if frame:
                velocities = (other_job.results.rkfpath(), frame)
                molecule = other_job.results.get_history_molecule(frame)
            else:
                velocities = other_job
                molecule = other_job.results.get_main_molecule()
        else:
            velocities = None
            molecule = None

        extra_settings = other_job.settings.copy()
        if settings:
            extra_settings.update(settings)

        if "InitialVelocities" in extra_settings.input.ams.MolecularDynamics:
            del extra_settings.input.ams.MolecularDynamics.InitialVelocities

        if "System" in extra_settings.input.ams:
            del extra_settings.input.ams.System

        return other_job, velocities, molecule, extra_settings

    def _get_thermostat_settings(self, thermostat, temperature, tau, thermostat_region, nsteps: int):
        s = Settings()
        prev_thermostat_settings = self.settings.input.ams.MolecularDynamics.Thermostat
        if isinstance(prev_thermostat_settings, list):
            prev_thermostat_settings = prev_thermostat_settings[0]
        s.input.ams.MolecularDynamics.Thermostat.Type = (
            thermostat or prev_thermostat_settings.Type or self.default_thermostat
        )
        try:
            n_temperatures = len(temperature)
            my_temperature = " ".join(str(x) for x in temperature)
            if n_temperatures > 1:
                nsteps_per_temperature = nsteps // (len(temperature) - 1)
                s.input.ams.MolecularDynamics.Thermostat.Duration = " ".join(
                    [str(nsteps_per_temperature)] * (n_temperatures - 1)
                )
            else:
                s.input.ams.MolecularDynamics.Thermostat.Duration = None
        except TypeError:
            my_temperature = temperature
            s.input.ams.MolecularDynamics.Thermostat.Duration = None

        s.input.ams.MolecularDynamics.Thermostat.Region = (
            thermostat_region or prev_thermostat_settings.get("Region", None) or self.default_thermostat_region
        )
        s.input.ams.MolecularDynamics.Thermostat.Temperature = (
            my_temperature or prev_thermostat_settings.Temperature or self.default_temperature
        )
        s.input.ams.MolecularDynamics.Thermostat.Tau = (
            tau
            or prev_thermostat_settings.Tau
            or float(self.settings.input.ams.MolecularDynamics.TimeStep) * AMSMDJob.default_tau_multiplier
        )
        return s

    def _get_barostat_settings(self, pressure, barostat, barostat_tau, scale, equal, constantvolume):
        s = Settings()
        self.settings.input.ams.MolecularDynamics.Barostat.Type = (
            barostat or self.settings.input.ams.MolecularDynamics.Barostat.Type or self.default_barostat
        )
        self.settings.input.ams.MolecularDynamics.Barostat.Pressure = (
            pressure
            if pressure is not None
            else self.settings.input.ams.MolecularDynamics.Barostat.Pressure or self.default_pressure
        )
        self.settings.input.ams.MolecularDynamics.Barostat.Tau = (
            barostat_tau
            or self.settings.input.ams.MolecularDynamics.Barostat.Tau
            or float(self.settings.input.ams.MolecularDynamics.TimeStep) * AMSMDJob.default_barostat_tau_multiplier
        )
        self.settings.input.ams.MolecularDynamics.Barostat.Scale = (
            scale or self.settings.input.ams.MolecularDynamics.Barostat.Scale or self.default_scale
        )
        self.settings.input.ams.MolecularDynamics.Barostat.Equal = (
            equal or self.settings.input.ams.MolecularDynamics.Barostat.Equal or self.default_equal
        )
        self.settings.input.ams.MolecularDynamics.Barostat.ConstantVolume = (
            str(constantvolume)
            if constantvolume is not None
            else self.settings.input.ams.MolecularDynamics.Barostat.ConstantVolume or self.default_constantvolume
        )
        return s

    @classmethod
    def restart_from(cls, other_job, frame=None, settings=None, use_prerun=False, **kwargs):
        """

        other_job: AMSJob
            The job to restart from.

        frame: int
            Which frame to read the structure and velocities from. If None, the final structure and end velocities will be used (section Molecule and MDResults%EndVelocities)

        settings: Settings
            Settings that override any other settings. All settings from other_job (e.g. the engine settings) are inherited by default but they can be overridden here.

        use_prerun: bool
            If True, the molecule and velocities will only be read from other_job inside the prerun() method. Set this to True to prevent PLAMS from waiting for other_job to finish as soon as the new job is defined.

        kwargs: many options
            See the docstring for AMSMDJob.

        """
        other_job, velocities, molecule, extra_settings = cls._get_restart_job_velocities_molecule(
            other_job, frame, settings, get_velocities_molecule=not use_prerun
        )
        job = cls(settings=extra_settings, velocities=velocities, molecule=molecule, **kwargs)

        if use_prerun:

            @add_to_instance(job)
            def prerun(self):  # noqa F811
                self.get_velocities_from(other_job, frame=frame, update_molecule=True)

        return job


class AMSNVEJob(AMSMDJob):
    """
    A class for running NVE MD simulations
    """

    def __init__(self, **kwargs):
        AMSMDJob.__init__(self, **kwargs)
        self.remove_blocks(["thermostat", "barostat", "deformation", "nanoreactor"])


class AMSNVTJob(AMSMDJob):
    """
    A class for running NVT MD simulations
    """

    def __init__(self, **kwargs):
        AMSMDJob.__init__(self, _enforce_thermostat=True, **kwargs)
        self.remove_blocks(["barostat", "deformation", "nanoreactor"])


class AMSNPTResults(AMSResults):
    def get_equilibrated_molecule(self, equilibration_fraction=0.667, return_index=False):
        """

        Discards the first equilibration_fraction of the trajectory.
        Calculates the average density of the rest. Returns the molecule
        with the closest density to the average density among the remaining
        trajectory.

        """
        densities = self.job.results.get_history_property("Density", "MDHistory")
        analyze_from = int(len(densities) * equilibration_fraction)
        # take structure closest to the target density
        avg_density = np.mean(densities[analyze_from:])  # amu/bohr^3
        delta = np.array(densities[analyze_from:]) - avg_density
        delta = np.abs(delta)
        min_index = np.argmin(delta)
        min_index += analyze_from
        mol = self.job.results.get_history_molecule(min_index + 1)

        if return_index:
            return mol, min_index + 1

        return mol


class AMSNPTJob(AMSMDJob):
    """
    A class for running NPT MD simulations
    """

    _result_type = AMSNPTResults

    def __init__(self, **kwargs):
        AMSMDJob.__init__(self, _enforce_thermostat=True, _enforce_barostat=True, **kwargs)
        self.settings.input.ams.MolecularDynamics.CalcPressure = "True"
        self.remove_blocks(["deformation", "nanoreactor"])



#!/usr/bin/env amspython
# coding: utf-8

# ## Initial imports

import scm.plams as plams
import matplotlib.pyplot as plt
import numpy as np

plams.init()


# ## Molecule

mol = plams.from_smiles("NC(CO)OCC=O")
plams.plot_molecule(mol)


# ## Engine settings

s = plams.Settings()
s.input.GFNFF  # GFN-FF force field
# s.input.MLPotential.Model = "AIMNet2-wB97MD3"   # new ml model in AMS2024

print(plams.AMSJob(settings=s).get_input())

# run in serial
s.runscript.nproc = 1


# ## Some general parameters

T = 298  # temperature in K
max_freq = 4000  # plot spectrum up to 4000 cm^-1
max_dt_fs = 2000  # maximum correlation in fs for dipole derivative acf


# ## Equilibration
#
# ``temperature=(500, T, T)`` means that in the first half the simulation the system is cooled from 500 K to the gievn temperature, and then kept constant at that temperature.
#
# The initial temperature of 500 K does some preliminary conformer search.

eq_job = plams.AMSNVTJob(
    settings=s,
    name="nvt_eq",
    molecule=mol,
    timestep=0.5,
    nsteps=10000,
    temperature=(500, T, T),
)
eq_job.run()


# ## NVE production simulation
#
# The ``binlog_dipolemoment`` option stores the dipole moment at every time step.

job = plams.AMSNVEJob.restart_from(
    eq_job,
    name="nve_single_prod",
    nsteps=50000,
    binlog_dipolemoment=True,
    binlog_time=True,
    samplingfreq=100,
    timestep=0.5,
)
job.run()


# ## Dipole derivative autocorrelation function

times, dipole_deriv_acf = job.results.get_dipole_derivatives_acf(start_fs=0, max_dt_fs=max_dt_fs)
plt.plot(times, dipole_deriv_acf)
plt.xlabel("Time (fs)")
plt.ylabel("Dipole deriv. autocorrelation (e bohr / fs)^2")
plt.title("Raw autocorrelation function")


# Ideally, you should set ``max_dt_fs`` above to a large enough number so that the autocorrelation function decreases to a constant value of 0 (**and** have a long enough MD simulation to get enough statistics!)

# ## IR spectrum
#
# The IR spectrum is the Fourier transform of the above autocorrelation function:

x_freq, y_intens_raw = job.results.get_ir_spectrum_md(times, dipole_deriv_acf, max_freq=max_freq)
plt.plot(x_freq, y_intens_raw)
plt.xlabel("Frequency (cm^-1)")
plt.title("IR spectrum (from raw autocorrelation function)")
plt.xlim(500, max_freq)
plt.show()


# There seems to be quite some "noise" in the IR spectrum. One reason for this is that there is still some signal (or noise?) in the autocorrelation function at dt = 2000 fs.
#
# However, it's also possible to use a tapering (window) function to make the autocorrelation function smoothly decrease to 0. This will make the resulting IR spectrum look a bit more tidy. See the next section.

# ## Tapering function for autocorrelation function


def tapered_cosine(x):
    return 0.5 * (np.cos(np.pi * x / np.max(x)) + 1)


plt.plot(times, tapered_cosine(times))
plt.title("Tapering / cutoff / window function")


# Now apply this function to the autocorrelation function:

dipole_deriv_acf_tapered_cosine = dipole_deriv_acf * tapered_cosine(times)
plt.plot(times, dipole_deriv_acf_tapered_cosine)
plt.xlabel("Time (fs)")
plt.title("Autocorrelation cosine tapering")


# And calculate the IR spectrum:

x_freq, y_intens_cosine = job.results.get_ir_spectrum_md(times, dipole_deriv_acf_tapered_cosine, max_freq=max_freq)
plt.plot(x_freq, y_intens_cosine)
plt.xlabel("Frequency (cm^-1)")
plt.title("IR spectrum (from cosine-tapered autocorrelation function)")
plt.xlim(500, max_freq)


# Above we see that using the cosine-tapered autocorrelation function gives a smoother IR spectrum without affecting the intensities too much.

# ## Compare to IR spectrum calculated from harmonic approximation

# Let's compare to an IR spectrum calculated with a geometry optimization + frequencies job, starting from the final frame of the MD simulation.

ams_s = plams.Settings()
ams_s.input.ams.Task = "GeometryOptimization"
ams_s.input.ams.Properties.NormalModes = "Yes"
harmonic_mol = job.results.get_main_molecule()
harmonic_job = plams.AMSJob(settings=ams_s + s, name="harmonic", molecule=harmonic_mol)
harmonic_job.run()


harmonic_freq, harmonic_intens = harmonic_job.results.get_ir_spectrum(broadening_type="lorentzian", broadening_width=20)
plt.plot(harmonic_freq, harmonic_intens)
rescale_factor = np.sum(harmonic_intens) / np.sum(y_intens_cosine)
plt.plot(x_freq, y_intens_cosine * rescale_factor)  # rescale
plt.legend(["Harmonic", "MD NVE"])
plt.title("IR spectrum")
plt.xlabel("Frequency (cm^-1)")
plt.xlim(500, max_freq)


# In this case, the MD simulation samples multiple conformers so there are more peaks than for the harmonic calculation.
#
# For example, the peak for the MD at 3600 cm^-1 corresponds to the "free" OH stretch of the hydroxyl group, but in conformer used for the  harmonic approximation the hydroxyl donates a hydrogen bond to the aldehyde oxygen (giving a lower vibrational frequency):

plams.plot_molecule(harmonic_job.results.get_main_molecule())


# ## View the trajectory in AMSmovie


# ## View the normal modes in AMSspectra


# ## Appendix: Average over multiple short NVE simulations
#
# Best practice is to run multiple NVE simulations starting from different points of the NVT simulation, assuming that the NVT simulation samples the correct equilibrium distribution of structures/conformers.
#
# Let's make this more explicit with another NVT simulation, followed by multiple NVE simulations from different points of the NVT simulation. See also the "Molecular Dynamics with Python" tutorial.

nvt_prod_job = plams.AMSNVTJob.restart_from(
    eq_job,
    name="nvt_prod",
    nsteps=50000,
    samplingfreq=100,
    timestep=0.5,
    thermostat="NHC",
    tau=100,
    temperature=T,
)
nvt_prod_job.run()


nvespawner_job = plams.AMSNVESpawnerJob(
    nvt_prod_job,
    name="nvespawner-" + nvt_prod_job.name,
    n_nve=10,  # the number of NVE simulations to run
    timestep=0.5,
    binlog_time=True,
    binlog_dipolemoment=True,
    nsteps=20000,
)
nvespawner_job.run()


# Let's check that the temperature during the NVE is not too far from the requested temperature.

avg_T = nvespawner_job.results.get_mean_temperature()

print(f"Set temperature during NVT: {T:.1f} K")
print(f"Mean temperature during NVE: {avg_T:.1f}")


# Calculate the average dipole derivative autocorrelation function.
#
# To calculate the IR spectrum from our custom set of averaged data, we directly call the ``power_spectrum`` function from PLAMS:

avg_x, avg_y = nvespawner_job.results.get_dipole_derivatives_acf(start_fs=0, max_dt_fs=max_dt_fs)
avg_y *= tapered_cosine(avg_x)

x_freq_multiple, y_intens_cosine_multiple = plams.trajectories.analysis.power_spectrum(times, avg_y, max_freq=max_freq)


plt.plot(x_freq, y_intens_cosine)
plt.plot(x_freq_multiple, y_intens_cosine_multiple)
plt.xlabel("Frequency (cm^-1)")
plt.title("IR spectrum from multiple NVE simulations")
plt.legend(["single", "multiple"])
plt.xlim(500, 4000)



import sys
from scm.plams import RKFTrajectoryFile


def main():
    """
    Main body of the script
    """
    filename = sys.argv[1]

    rkf = RKFTrajectoryFile(filename)
    mol = rkf.get_plamsmol()
    rkf.close()

    oxygens = get_water_oxygens(mol)

    outfile = open("indices.txt", "w")
    for i, io in enumerate(oxygens):
        if i % 10 == 0 and i != 0:
            outfile.write("\n")
        outfile.write("%8i " % (io))
    outfile.close()


def get_water_oxygens(mol):
    """
    Select the oxygens in water only
    """
    wo = []
    for iat, at in enumerate(mol.atoms):
        if at.symbol != "O":
            continue
        neighbors = [n.symbol for n in mol.neighbors(at)]
        if neighbors == ["H", "H"]:
            wo.append(iat)
    return wo


if __name__ == "__main__":
    main()



from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults

__all__ = ["UFFJob", "UFFResults"]


class UFFResults(SCMResults):
    _kfext = ".rkf"
    _rename_map = {"uff.rkf": "$JN" + _kfext}


class UFFJob(SCMJob):
    _result_type = UFFResults
    _command = "uff"
    _top = ["title", "units"]
    _subblock_end = "end"

    def _serialize_mol(self):
        s = self.settings.input

        printtypes = all(map(lambda at: ("uff" in at.properties and "type" in at.properties.uff), self.molecule))

        for i, atom in enumerate(self.molecule):
            suffix = ""
            suffix_dict = {}
            if printtypes:
                suffix = "  {type}  {charge}"
                suffix_dict = atom.properties.uff.copy()
                if "charge" not in suffix_dict:
                    suffix_dict.charge = 0.0
            s.system.atoms["_" + str(i + 1)] = atom.str(suffix=suffix, suffix_dict=suffix_dict, space=18, decimal=10)

        if "ignore_bonds" not in self.settings:
            self.molecule.set_atoms_id()
            for i, bond in enumerate(self.molecule.bonds):
                s.system.bonds["_" + str(i + 1)] = "{:6d}  {:6d}  {:6.2f}".format(
                    bond.atom1.id, bond.atom2.id, bond.order
                )
            self.molecule.unset_atoms_id()

        if self.molecule.lattice:
            for i, vec in enumerate(self.molecule.lattice):
                s.system.lattice["_" + str(i + 1)] = "{:16.10f} {:16.10f} {:16.10f}".format(*vec)

    def _remove_mol(self):
        s = self.settings.input

        if "system" in s:
            for k in ["atoms", "bonds", "lattice"]:
                if k in s.system:
                    del s.system[k]



#!/usr/bin/env amspython
import os
import sys
from collections import Counter

import matplotlib.pyplot as plt
import numpy as np
from scm.plams import *

"""

Example printing a table of the AMS Molecule analysis and/or AMS Bond analysis
from reactive MD simulations.

For molecules, for each species and frame, print the number of species in that
frame to molecule_analysis.txt and the corresponding plot in
molecule_analysis.pdf.

For bonds, for each frame, print the number of bonds per type in that frame to
bond_analysis.txt and the corresponding plot to bond_analysis.pdf.  At the top
of the main() function, set whether to round the bond orders to the nearest
integer or half-integer.

You can comment out the molecule/bonds/print/plot functions to fit your needs.

Usage: $AMSBIN/amspython MoleculesTable.py /path/to/ams.rkf

If no /path/to/ams.rkf is provided, the script will run a short toy MD with
ReaxFF to illustrate the analysis (requires a ReaxFF license).

"""


def main():
    round_type = "integer"  # round bond orders to nearest integer
    # round_type = 'half_integer' # round bond orders to nearest half integer

    if len(sys.argv) > 2:
        print("Usage: $AMSBIN/amspython MoleculesTable.py /path/to/ams.rkf")
        print("   or: $AMSBIN/amspython MoleculesTable.py")
        exit(1)

    elif len(sys.argv) == 2:
        ams_rkf_path = sys.argv[1]

    elif len(sys.argv) == 1:
        # Run a short combustion MD with ReaxFF
        ams_rkf_path = run_md()

    try:
        # Molecule analysis
        names, counts_list = analyze_molecules(ams_rkf_path)
        molecule_analysis_table = table_results(names, counts_list)
        with open("molecule_analysis.txt", "w") as f:
            f.write(molecule_analysis_table)
        plot_results(names, counts_list, "molecule_analysis.pdf")

        # Bond analysis
        names, counts_list = analyze_bonds(ams_rkf_path, round_type)
        bond_analysis_table = table_results(names, counts_list)
        with open("bond_analysis.txt", "w") as f:
            f.write(bond_analysis_table)
        plot_results(names, counts_list, "bond_analysis.pdf")

    except Exception as e:
        print(e)
        exit(1)


def table_results(names, counts_list):
    """
    This function returns the results in a space-separated table.

    names: list of str
        Molecule names

    counts_list: list of list of int
        List with dimensions nFrames x nMoleculeTypes
    """
    ret = ["frame " + " ".join(names)]

    for frame, counts in enumerate(counts_list, 1):
        line = f"{frame} " + " ".join(str(x) for x in counts)
        ret.append(line)

    return "\n".join(ret)


def plot_results(names, counts_list, filename=None):
    """
    Plot the results as Count vs. frame
    """
    data = np.array(counts_list).T
    for i, name in enumerate(names, 0):
        plt.plot(data[i], label=name)
    plt.legend()
    plt.xlabel("Frame")
    plt.ylabel("Count")
    plt.tight_layout()
    if filename:
        plt.savefig(filename)
    plt.show()


def round_off(number, rtype="integer"):
    """
    Round off bond order to integer or half integer.
    """
    if rtype == "half_integer":
        number = round(number * 2) / 2
    if rtype == "integer":
        number = round(number)
    return number


def bo2symbol(bo):
    """
    Convert bond order to bond symbol.
    """
    s = "---"
    if bo == 0.5:
        s = "--"
    if bo == 1.0:
        s = "-"
    if bo == 1.5:
        s = "=="
    if bo == 2.0:
        s = "="
    if bo == 2.5:
        s = "≡≡"
    if bo == 3.0:
        s = "≡"
    return s


def analyze_molecules(ams_rkf_path):
    """
    ams_rkf_path: str
        Path to an ams.rkf file from a reactive MD simulation

    Returns: 2-tuple (names, counts)
        ``names``: list of length nMoleculesTypes. ``counts``: list of
        length nFrames, each item a list of length nMoleculesTypes
        containing an integer with the number of molecules of that type at
        the particular frame.

    """

    if not os.path.exists(ams_rkf_path):
        raise FileNotFoundError(f"Couldn't find the file {ams_rkf_path}")

    job = AMSJob.load_external(ams_rkf_path)

    try:
        n_molecules = job.results.readrkf("Molecules", "Num molecules")
    except KeyError:
        raise KeyError(
            "Couldn't find Molecules section on ams.rkf. You need to enable MolecularDynamics%Trajectory%WriteMolecules (before running the MD simulation)."
        )

    # get the names of the molecules (molecular formula)
    molecule_type_range = range(1, n_molecules + 1)  # 1, 2, ..., n_molecules
    names = [job.results.readrkf("Molecules", f"Molecule name {i}") for i in molecule_type_range]

    # read the Mols.Type from each History element
    mols_type_list = job.results.get_history_property("Mols.Type")  # list of length nFrames
    counts_list = []

    # loop over the frames
    # store the counts-per-molecule-type in counts_list
    for mols_types in mols_type_list:
        counts = Counter(mols_types)
        counts_list.append([counts[x] for x in molecule_type_range])

    return names, counts_list


def analyze_bonds(ams_rkf_path, round_type):
    """
    ams_rkf_path: str
        Path to an ams.rkf file from a reactive MD simulation

    Returns: dict (number of bond type)
        Each element of the dict are a list of length nframes containing
        the occurence of every bond type at a given frame.
    """

    if not os.path.exists(ams_rkf_path):
        raise FileNotFoundError(f"Couldn't find the file {ams_rkf_path}")

    job = AMSJob.load_external(ams_rkf_path)

    trajectory = Trajectory(job.results.rkfpath())
    # Loop over frames
    allbonds = {}
    nframes = len(trajectory)
    for iframe, imol in enumerate(trajectory, 1):

        bonds, btype = [], []
        # mol.bonds double count bonds
        for bond in imol.bonds:

            # Bond indices and symbols
            b_atoms = [bond.atom1, bond.atom2]
            b_symbols = [bond.atom1.symbol, bond.atom2.symbol]
            sorted_symbols = sorted(b_symbols)

            # If bond has not yet been counted
            if b_atoms not in bonds:
                # Compute BO and add bond to list
                sorted_symbols.append(round_off(bond.order, round_type))
                btype.append(sorted_symbols)
                bonds.append(b_atoms)
                bonds.append([b_atoms[1], b_atoms[0]])

        # Loop over unique bonds
        btype_set = set(tuple(row) for row in btype)
        for ibtype in btype_set:
            # Count how many bond of given type
            n_ibond = btype.count(list(ibtype))
            b_tag = ibtype[0] + bo2symbol(ibtype[2]) + ibtype[1]
            b_label = "{:s} (BO {:2.1f})".format(b_tag, ibtype[-1])
            # If new bond at frame iframe then create zeros
            if b_label not in allbonds:
                allbonds[b_label] = [0.0] * (nframes)
            allbonds[b_label][iframe - 1] = n_ibond

        # Counter
        if iframe % 10 == 1:
            print("Currently at frame {:d}/{:d}".format(iframe, nframes))

    # Convert dictionary to counts_list
    names, counts_list = [name for name in allbonds], []
    for iframe, imol in enumerate(trajectory, 1):
        counts_list_iframe = []
        for name in names:
            counts_list_iframe.append(allbonds[name][iframe - 1])
        counts_list.append(counts_list_iframe)

    return names, counts_list


def run_md():
    """
    Define the simulation box as mixture of H2/O2 and run a short NVE at high T
    """

    init()
    o2 = from_smiles("O=O")
    h2 = from_smiles("[HH]")
    mixture = packmol(molecules=[o2, h2], n_molecules=[4, 4], density=1.0)

    s = Settings()
    s.input.reaxff.forcefield = "CHO.ff"
    s.input.ams.task = "MolecularDynamics"
    s.input.ams.MolecularDynamics.Nsteps = 3000
    s.input.ams.MolecularDynamics.Timestep = 0.5
    s.input.ams.MolecularDynamics.InitialVelocities.Temperature = 3500

    job = AMSJob(settings=s, molecule=mixture, name="md")
    job.run(watch=True)
    finish()

    return job.results.rkfpath()


if __name__ == "__main__":
    main()



import sys
from io import StringIO
import numpy
from scm.plams import RKFTrajectoryFile
from scm.flexmd import PDBMolecule


def main():
    """
    Main sctipt
    """
    if sys.argv[1] == "-h":
        raise SystemExit("amspython get_hbonds.py path/to/ams.rkf path/to/indices.txt")

    # Deal with the input
    filename = sys.argv[1]
    indexfile = sys.argv[2]
    indices, elements = read_indices(indexfile)

    # Open the trajectory file
    rkf = RKFTrajectoryFile(filename)
    mol = rkf.get_plamsmol()
    nsteps = len(rkf)
    print("NSteps: ", nsteps)
    for i, at in enumerate(mol.atoms):
        if at.symbol in elements:
            if not i in indices:
                indices.append(i)

    # Create the PDB object
    print("Creating the PDB object")
    f = StringIO()
    mol.writexyz(f)
    f.seek(0)
    text = f.read()
    f.close()
    pdb = PDBMolecule(xyzstring=text, bonddetectorflag=False)
    print("PDB created")

    # Get the actual number of HBonds per selected atom
    heavy_atoms = [i for i, at in enumerate(mol.atoms) if not at.symbol == "H"]
    hydrogens = [i for i, at in enumerate(mol.atoms) if at.symbol == "H"]
    values = []
    print("%8s %8s %s" % ("Step", "Atom", "Neighbors"))
    for istep in range(nsteps):
        crd, cell = rkf.read_frame(istep, molecule=mol)
        # Create neighborlists
        d_indices, boxlist = pdb.divide_into_cubes(range(len(mol)))
        pdb.coords = crd
        pdb.set_cellvectors(cell)
        for iat in indices:
            atomlists = (heavy_atoms, hydrogens)
            atoms, hs = pdb.find_neighbours_using_cubes(iat, d_indices, boxlist, atomlists)
            hbonds = pdb.get_hbonds(iat, atoms, hs)
            print("%8i %8i %s" % (istep, iat, str(hbonds)))
            values.append(len(hbonds))

    # Compute the histogram
    bins = [i for i in range(max(values) + 1)]
    yvalues, xvalues = numpy.histogram(values, bins=bins)

    # Write to output
    outfile = open("hist.txt", "w")
    for x, y in zip(xvalues, yvalues):
        outfile.write("%20.10f %20.10f\n" % (x, y / nsteps))
    outfile.close()


def read_indices(indexfilename):
    """
    Read atom indices from a file
    """
    infile = open(indexfilename)
    lines = infile.readlines()
    infile.close()
    indices = []
    elements = set()
    for line in lines:
        words = line.split()
        if len(words) == 0:
            continue
        digits = [w.isdigit() for w in words]
        for w in words:
            if w.isdigit():
                indices.append(int(w))
            else:
                elements.add(w)
    return indices, elements


if __name__ == "__main__":
    main()



from scm.plams import Settings
from scm.plams.interfaces.adfsuite.ase_calculator import AMSCalculator


def test_Properties():
    s = Settings()
    s.input.ForceField
    s.input.ams.Task = "SinglePoint"
    s.runscript.nproc = 1
    job = AMSCalculator(s, name="Properties")
    assert "forces" not in job.implemented_properties
    job.ensure_property("forces")
    assert "forces" in job.implemented_properties


if __name__ == "__main__":
    test_Properties()



from pathlib import Path
import pytest

from scm.plams import Molecule

PATH = Path(".") / "xyz"


@pytest.fixture
def m1(xyz_folder):
    return Molecule(xyz_folder / "CO_6_1.xyz")


@pytest.fixture
def m2(xyz_folder):
    return Molecule(xyz_folder / "CO_6_2.xyz")


@pytest.fixture
def n1(xyz_folder):
    return Molecule(xyz_folder / "CO_6_3.xyz")


@pytest.fixture
def n2(xyz_folder):
    return Molecule(xyz_folder / "CO_6_4.xyz")


@pytest.fixture
def n3(xyz_folder):
    return Molecule(xyz_folder / "CO_6_5.xyz")


def testYES(m1, m2, n1, n2, n3):
    for i in range(4):
        assert m1.label(i) == m2.label(i)
    for i in range(4):
        assert n1.label(i) == n2.label(i) == n3.label(i)


def testNO(m1, m2, n1, n2, n3):
    assert m1.label(4) != m2.label(4)
    assert n1.label(4) != n2.label(4)
    assert n1.label(4) != n3.label(4)
    assert n2.label(4) != n3.label(4)



from collections import OrderedDict

from scm.plams.core.basejob import MultiJob
from scm.plams.core.functions import add_to_instance
from scm.plams.core.results import Results
from scm.plams.interfaces.adfsuite.ams import AMSJob

__all__ = ["ReorganizationEnergyJob", "ReorganizationEnergyResults"]


# using this function to pass a molecule inside a MultiJob ensures proper parallel execution
def pass_molecule(source, target):
    @add_to_instance(target)
    def prerun(self):  # noqa F811
        self.molecule = source.results.get_main_molecule()


class ReorganizationEnergyResults(Results):
    """Results class for reorganization energy."""

    def reorganization_energy(self, unit="au"):
        energies = self.get_all_energies(unit)

        reorganization_energy = (
            energies["state B geo A"]
            - energies["state A geo A"]
            + energies["state A geo B"]
            - energies["state B geo B"]
        )
        return reorganization_energy

    def get_all_energies(self, unit="au"):
        energies = {}
        energies["state A geo A"] = self.job.children["go_A"].results.get_energy(unit=unit)
        energies["state B geo B"] = self.job.children["go_B"].results.get_energy(unit=unit)
        energies["state A geo B"] = self.job.children["sp_A_for_geo_B"].results.get_energy(unit=unit)
        energies["state B geo A"] = self.job.children["sp_B_for_geo_A"].results.get_energy(unit=unit)

        return energies


class ReorganizationEnergyJob(MultiJob):
    """A class for calculating the reorganization energy using AMS.
    Given two states, A and B, the reorganization energy is defined as follows:

    reorganization energy =
    E(state B at optimal geometry for state A) -
    E(state A at optimal geometry for state A) +
    E(state A at optimal geometry for state B) -
    E(state B at optimal geometry for state B)

    This job will run two geometry optimizations and two single point calculations.
    """

    _result_type = ReorganizationEnergyResults

    def __init__(self, molecule, common_settings, settings_state_A, settings_state_B, **kwargs):
        """
        molecule: the molecule
        common_settings: a setting object for an AMSJob containing the shared settings for all the calculations
        settings_state_A: Setting object for an AMSJob containing exclusivelt the options defining the state A (e.g. charge and spin)
        settings_state_B: Setting object for an AMSJob containing exclusivelt the options defining the state B (e.g. charge and spin)
        kwargs: other options to be passed to the MultiJob constructor
        """
        MultiJob.__init__(self, children=OrderedDict(), **kwargs)

        go_settings = common_settings.copy()
        go_settings.input.ams.task = "GeometryOptimization"

        sp_settings = common_settings.copy()
        sp_settings.input.ams.task = "SinglePoint"

        # copy the settings so that we wont modify the ones provided as input by the user
        settings_state_A = settings_state_A.copy()
        settings_state_B = settings_state_B.copy()
        # In case the charge key is not specified, excplicitely set the value to 0.
        # This is to prevent the charge in molecule.properties.charge (set by get_main_molecule())
        # to be used in case of neutral systems
        for s in [settings_state_A, settings_state_B]:
            if not "charge" in s.input.ams.system:
                s.input.ams.system.charge = 0

        self.children["go_A"] = AMSJob(molecule=molecule, settings=go_settings + settings_state_A, name="go_A")
        self.children["go_B"] = AMSJob(molecule=molecule, settings=go_settings + settings_state_B, name="go_B")
        self.children["sp_A_for_geo_B"] = AMSJob(settings=sp_settings + settings_state_A, name="sp_A_geo_B")
        self.children["sp_B_for_geo_A"] = AMSJob(settings=sp_settings + settings_state_B, name="sp_B_geo_A")
        pass_molecule(self.children["go_A"], self.children["sp_B_for_geo_A"])
        pass_molecule(self.children["go_B"], self.children["sp_A_for_geo_B"])



from scm.plams.mol.molecule import Molecule
from scm.plams.interfaces.adfsuite.ams import AMSJob
from typing import Tuple, Union, List

__all__ = ["plot_band_structure", "plot_molecule", "plot_correlation"]


def plot_band_structure(x, y_spin_up, y_spin_down=None, labels=None, fermi_energy=None, zero=None, show=False):
    """
    Plots an electronic band structure from DFTB or BAND with matplotlib.

    To control the appearance of the plot you need to call ``plt.ylim(bottom, top)``, ``plt.title(title)``, etc.
    manually outside this function.

    x: list of float
        Returned by AMSResults.get_band_structure()

    y_spin_up: 2D numpy array of float
        Returned by AMSResults.get_band_structure()

    y_spin_down: 2D numpy array of float. If None, the spin down bands are not plotted.
        Returned by AMSResults.get_band_structure()

    labels: list of str
        Returned by AMSResults.get_band_structure()

    fermi_energy: float
        Returned by AMSResults.get_band_structure(). Should have the same unit as ``y``.

    zero: None or float or one of 'fermi', 'vbmax', 'cbmin'
        Shift the curves so that y=0 is at the specified value. If None, no shift is performed. 'fermi', 'vbmax', and 'cbmin' require that the ``fermi_energy`` is not None. Note: 'vbmax' and 'cbmin' calculate the zero as the highest (lowest) eigenvalue smaller (greater) than or equal to ``fermi_energy``. This is NOT necessarily equal to the valence band maximum or conduction band minimum as calculated by the compute engine.

    show: bool
        If True, call plt.show() at the end
    """
    import matplotlib.pyplot as plt
    import numpy as np

    if zero is None:
        zero = 0
    elif zero == "fermi":
        assert fermi_energy is not None
        zero = fermi_energy
    elif zero in ["vbm", "vbmax"]:
        assert fermi_energy is not None
        zero = y_spin_up[y_spin_up <= fermi_energy].max()
        if y_spin_down is not None:
            zero = max(zero, y_spin_down[y_spin_down <= fermi_energy].max())
    elif zero in ["cbm", "cbmax"]:
        assert fermi_energy is not None
        zero = y_spin_up[y_spin_up >= fermi_energy].min()
        if y_spin_down is not None:
            zero = min(zero, y_spin_down[y_spin_down <= fermi_energy].min())

    labels = labels or []

    fig, ax = plt.subplots()

    plt.plot(x, y_spin_up - zero, "-")
    if y_spin_down is not None:
        plt.plot(x, y_spin_down - zero, "--")

    tick_x = []
    tick_labels = []
    for xx, ll in zip(x, labels):
        if ll:
            if len(tick_x) == 0:
                tick_x.append(xx)
                tick_labels.append(ll)
                continue
            if np.isclose(xx, tick_x[-1]):
                if ll != tick_labels[-1]:
                    tick_labels[-1] += f",{ll}"
            else:
                tick_x.append(xx)
                tick_labels.append(ll)

    for xx in tick_x:
        plt.axvline(xx)

    if fermi_energy is not None:
        plt.axhline(fermi_energy - zero, linestyle="--")

    plt.xticks(ticks=tick_x, labels=tick_labels)

    if show:
        plt.show()


def plot_molecule(molecule, figsize=None, ax=None, keep_axis: bool = False, **kwargs):
    """Show a molecule in a Jupyter notebook"""
    import matplotlib.pyplot as plt
    from ase.visualize.plot import plot_atoms
    from scm.plams.interfaces.molecule.ase import toASE

    if isinstance(molecule, Molecule):
        molecule = toASE(molecule)

    if not ax:
        plt.figure(figsize=figsize or (2, 2))

    plot_atoms(molecule, ax=ax, **kwargs)

    if not keep_axis:
        if ax:
            ax.axis("off")
        else:
            plt.axis("off")


def get_correlation_xy(
    job1: Union[AMSJob, List[AMSJob]],
    job2: Union[AMSJob, List[AMSJob]],
    section: str,
    variable: str,
    alt_section: str = None,
    alt_variable: str = None,
    file: str = "ams",
    multiplier: float = 1.0,
) -> Tuple:
    import numpy as np

    def tolist(x):
        if isinstance(x, list):
            return x
        return [x]

    job1 = tolist(job1)
    job2 = tolist(job2)

    alt_section = alt_section or section
    alt_variable = alt_variable or variable

    data1 = []
    data2 = []
    for j1, j2 in zip(job1, job2):
        try:
            d1 = j1.results.readrkf(section, variable, file=file)
        except KeyError:
            d1 = j1.results.get_history_property(variable, history_section=section)
        d1 = np.ravel(d1) * multiplier

        try:
            d2 = j2.results.readrkf(alt_section, alt_variable, file=file)
        except KeyError:
            d2 = j2.results.get_history_property(alt_variable, history_section=alt_section)
        d2 = np.ravel(d2) * multiplier

        data1.extend(list(d1))
        data2.extend(list(d2))

    data1 = np.array(data1)
    data2 = np.array(data2)

    return data1, data2


def plot_correlation(
    job1: Union[AMSJob, List[AMSJob]],
    job2: Union[AMSJob, List[AMSJob]],
    section: str,
    variable: str,
    alt_section: str = None,
    alt_variable: str = None,
    file: str = "ams",
    multiplier: float = 1.0,
    unit: str = None,
    save_txt: bool = None,
    ax=None,
    show_xy: bool = True,
    show_linear_fit: bool = True,
    show_mad: bool = True,
    show_rmsd: bool = True,
    xlabel: str = None,
    ylabel: str = None,
):
    """

    Plot a correlation plot from AMS .rkf files

    job1: AMSJob or List[AMSJob]
        Job(s) plotted on x-axis

    job2: AMSJob or List[AMSJob]
        job2: Job(s) plotted on y-axis

    section: str
        section: section to read on .rkf files

    variable: str
        variable: variable to read

    alt_section: str
        Section to read on .rkf files for job2. If not specified it will be the same as ``section``

    alt_variable : str
        Variable to read for job2. If not specified it will be the same as ``variable``.

    file: str, optional
        file: "ams" or "engine", defaults to "ams"

    multiplier: float, optional
        multiplier: Numbers will be multiplied by this number, defaults to 1.0

    unit: str, optional
        unit: unit will be shown in the plot, defaults to None

    save_txt: str, optional
        save_txt: If not None, save the xy data to this text file, defaults to None

    ax: matplotlib axis, optional
        ax: matplotlib axis, defaults to None

    show_xy: bool, optional
        show_xy: Whether to show y=x line, defaults to True

    show_linear_fit: bool, optional
        show_linear_fit: Whether to perform and show a linear fit, defaults to True

    show_mad: bool, optional
        show_mad: Whether to show mean absolute deviation, defaults to True

    show_rmsd: bool, optional
        show_rmsd: Whether to show root-mean-square deviation, defaults to True

    xlabel: str, optional
        xlabel: The x-label. If not given will be a list of job names, defaults to None

    ylabel: str, optional
        ylabel: THe y-label. If not given will be al ist of job names, defaults to None

    Returns: A matplotlib axis

    """

    import matplotlib.pyplot as plt
    import numpy as np

    def tolist(x):
        if isinstance(x, list):
            return x
        return [x]

    job1 = tolist(job1)
    job2 = tolist(job2)

    alt_section = alt_section or section
    alt_variable = alt_variable or variable

    data1, data2 = get_correlation_xy(job1, job2, section, variable, alt_section, alt_variable, file, multiplier)

    def add_unit(s: str):
        if unit is not None:
            return f"{s} ({unit})"

        return s

    if ax is None:
        fig, ax = plt.subplots()

    complete_data = np.stack((data1, data2), axis=1)

    min_data = np.min(complete_data)
    max_data = np.max(complete_data)
    min_max = np.array([min_data, max_data])

    legend = []
    title = [f"{section}%{variable}"]
    if show_xy:
        ax.plot(min_max, min_max, "-")
        legend.append("y=x")

    stats_title = ""

    if show_mad:
        mad = np.mean(np.abs(data2 - data1))
        stats_title += add_unit(f" MAD: {mad:.5f}")

    if show_rmsd:
        rmsd = np.sqrt(np.mean((data2 - data1) ** 2))
        stats_title += add_unit(f" RMSD: {rmsd:.5f}")

    linear_fit_title = None
    if show_linear_fit:
        from scipy.stats import linregress

        result = linregress(data1, data2)
        min_max_linear_fit = result.slope * min_max + result.intercept
        r2 = result.rvalue**2
        ax.plot(min_max, min_max_linear_fit, "-")
        legend.append("Fit")
        stats_title += f" R^2: {r2:.3f}"
        linear_fit_title = f"Linear fit slope={result.slope:.3f} intercept={result.intercept:.3f}"

    if stats_title:
        title.append(stats_title)

    if linear_fit_title:
        title.append(linear_fit_title)

    ax.plot(data1, data2, ".")
    legend.append("data")

    if xlabel is None:
        xlabel = ", ".join(x.name for x in job1)
        if len(xlabel) > 40:
            xlabel = xlabel[:35] + "..."
    if ylabel is None:
        ylabel = ", ".join(x.name for x in job2)
        if len(ylabel) > 40:
            ylabel = ylabel[:35] + "..."

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    ax.set_title("\n".join(title))
    ax.legend(legend)

    ax.set_box_aspect(1)
    ax.set_xlim(*min_max)
    ax.set_ylim(*min_max)

    if save_txt is not None:
        np.savetxt(save_txt, complete_data, header=f"{xlabel} {ylabel}")

    return ax


def plot_msd(job, start_time_fit_fs=None, ax=None):
    """
    job: AMSMSDJob
        The job for which to plot the results

    start_time_fit_fs: float
        The start time (in fs) for which to perform the linear fit

    ax: matplotlib axis
        The axis. If None, one will be created

    Returns: matplotlib axis
    """
    import matplotlib.pyplot as plt

    if ax is None:
        _, ax = plt.subplots()

    time, msd = job.results.get_msd()
    fit_result, fit_x, fit_y = job.results.get_linear_fit(start_time_fit_fs=start_time_fit_fs)
    # the diffusion coefficient can also be calculated as fit_result.slope/6 (ang^2/fs)
    diffusion_coefficient = job.results.get_diffusion_coefficient(start_time_fit_fs=start_time_fit_fs)  # m^2/s
    ax.plot(time, msd, label="MSD")
    ax.plot(fit_x, fit_y, label="Linear fit slope={:.5f} ang^2/fs".format(fit_result.slope))
    ax.legend()
    ax.set_xlabel("Correlation time (fs)")
    ax.set_ylabel("Mean square displacement (ang^2)")
    ax.set_title("MSD: Diffusion coefficient = {:.2e} m^2/s".format(diffusion_coefficient))

    return ax



#!/usr/bin/env amspython
from scm.plams import *

""" Run as: $AMSBIN/amspython BAND_NiO_HubbardU.py """


def main():
    d = 2.085
    mol = Molecule()
    mol.add_atom(Atom(symbol="Ni", coords=(0, 0, 0)))
    mol.add_atom(Atom(symbol="O", coords=(d, d, d)))
    mol.lattice = [[0.0, d, d], [d, 0.0, d], [d, d, 0.0]]

    s = Settings()
    s.input.ams.task = "SinglePoint"
    s.input.band.Unrestricted = "yes"
    s.input.band.XC.GGA = "BP86"
    s.input.band.Basis.Type = "DZP"
    s.input.band.KSpace.Quality = "Basic"
    s.input.band.NumericalQuality = "Normal"
    s.input.band.DOS.CalcPDOS = "Yes"
    s.input.band.HubbardU.Enabled = "Yes"
    s.input.band.HubbardU.UValue = "0.6 0.0"
    s.input.band.HubbardU.LValue = "2 -1"

    job = AMSJob(settings=s, molecule=mol, name="NiO")
    job.run()

    toeV = Units.convert(1.0, "hartree", "eV")
    topvb = job.results.readrkf("BandStructure", "TopValenceBand", file="engine") * toeV
    bottomcb = job.results.readrkf("BandStructure", "BottomConductionBand", file="engine") * toeV
    gap = bottomcb - topvb

    log("Results:")
    log(f"Top of valence band:       {topvb:.2f} eV")
    log(f"Bottom of conduction band: {bottomcb:.2f} eV")
    log(f"Band gap:                  {gap:.2f} eV")


if __name__ == "__main__":
    init()
    try:
        main()
    finally:
        finish()



import sys
from scm.plams import RKFTrajectoryFile, DCDTrajectoryFile
from scm.flexmd import pdb_from_plamsmol, PSFTopology


def main():
    """
    Main script
    """
    rkf = RKFTrajectoryFile("ams.rkf")
    mol = rkf.get_plamsmol()
    print("NSTeps: ", len(rkf))

    pdb = pdb_from_plamsmol(mol)
    psf = PSFTopology(pdb=pdb)
    psf.write_psf("ams.psf")

    dcd = DCDTrajectoryFile("ams.dcd", mode="wb")

    for i in range(len(rkf)):
        if i % 100 == 0:
            print(i)
        crd, cell = rkf.read_frame(i)
        mol.from_array(crd)
        mol.map_atoms_to_bonds()
        dcd.write_next(coords=mol.as_array(), cell=cell)


if __name__ == "__main__":
    main()
